"""
PC-BASIC - console.py
Console and interactive environment

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging

from ..compat import iterchar, int2byte

from .base import error
from .base import tokens as tk
from .base import scancode
from .base.eascii import as_bytes as ea


# alt+key macros for interactive mode
# these happen at a higher level (in the console) than F-key macros (in the keyboard buffer)
ALT_KEY_REPLACE = {
    ea.ALT_a: tk.KW_AUTO,
    ea.ALT_b: tk.KW_BSAVE,
    ea.ALT_c: tk.KW_COLOR,
    ea.ALT_d: tk.KW_DELETE,
    ea.ALT_e: tk.KW_ELSE,
    ea.ALT_f: tk.KW_FOR,
    ea.ALT_g: tk.KW_GOTO,
    ea.ALT_h: tk.KW_HEX,
    ea.ALT_i: tk.KW_INPUT,
    ea.ALT_k: tk.KW_KEY,
    ea.ALT_l: tk.KW_LOCATE,
    ea.ALT_m: tk.KW_MOTOR,
    ea.ALT_n: tk.KW_NEXT,
    ea.ALT_o: tk.KW_OPEN,
    ea.ALT_p: tk.KW_PRINT,
    ea.ALT_r: tk.KW_RUN,
    ea.ALT_s: tk.KW_SCREEN,
    ea.ALT_t: tk.KW_THEN,
    ea.ALT_u: tk.KW_USING,
    ea.ALT_v: tk.KW_VAL,
    ea.ALT_w: tk.KW_WIDTH,
    ea.ALT_x: tk.KW_XOR,
}


# characters to replace in bottom bar
FKEY_MACRO_REPLACE_CHARS = {
    b'\x07': b'\x0e', b'\x08': b'\xfe', b'\x09': b'\x1a', b'\x0A': b'\x1b',
    b'\x0B': b'\x7f', b'\x0C': b'\x16', b'\x0D': b'\x1b', b'\x1C': b'\x10',
    b'\x1D': b'\x11', b'\x1E': b'\x18', b'\x1F': b'\x19'
}


class Console(object):
    """Console / interactive environment."""

    def __init__(self, text_screen, cursor, keyboard, sound, io_streams, num_fn_keys, tandy_fn_keys):
        """Initialise environment."""
        self._text_screen = text_screen
        self._cursor = cursor
        self._sound = sound
        self._keyboard = keyboard
        self._io_streams = io_streams
        self._num_fn_keys = num_fn_keys
        self._tandy_fn_keys = tandy_fn_keys
        # overwrite mode (instead of insert)
        self._overwrite_mode = True
        # needs to be set later due to init order
        self._lpt1_file = None
        self._update_bar()

    def set_lpt1_file(self, lpt1_file):
        """Set the LPT1: file."""
        self._lpt1_file = lpt1_file

    ##########################################################################
    # properties

    @property
    def width(self):
        """Number of columns."""
        return self._text_screen.mode.width

    @property
    def height(self):
        """Number of rows."""
        return self._text_screen.mode.height

    @property
    def current_row(self):
        """Cursor row."""
        return self._text_screen.current_row

    @property
    def current_col(self):
        """Cursor column."""
        return self._text_screen.current_col

    @property
    def overflow(self):
        """Cursor is to the right of rightmost row."""
        return self._text_screen.overflow

    def set_pos(self, row, col):
        """Set cursor position."""
        self._text_screen.set_pos(row, col)

    ##########################################################################
    # interaction

    def read_line(self, prompt=b'', write_endl=True, is_input=False):
        """Enter interactive mode and read string from console."""
        self.write(prompt)
        # disconnect the wrap between line with the prompt and previous line
        if self._text_screen.current_row > 1:
            self._text_screen.set_wrap(self._text_screen.current_row-1, False)
        # read from start in direct entry mode, from prompt in input mode
        prompt_width = 0 if not is_input else self._text_screen.current_col - 1
        try:
            # give control to user for interactive mode
            prompt_row, left, right = self._interact(prompt_width, is_input=is_input)
        except error.Break:
            # x0E CR LF is printed to redirects at break
            self._io_streams.write(b'\x0e')
            # while only a line break appears on the console
            self.write_line()
            raise
        # get contents of the logical line
        if not is_input:
            left, right = 1, None
        else:
            # we need *inclusive* bounds
            right -= 1
        outstr = self._text_screen.get_logical_line(
            self._text_screen.current_row, furthest_left=left, furthest_right=right,
        )
        # redirects output exactly the contents of the logical line
        # including any trailing whitespace and chars past 255
        self._io_streams.write(outstr)
        # go to last row of logical line
        self._text_screen.move_to_end()
        # echo the CR, if requested
        if write_endl:
            self.write_line()
        # to the parser/INPUT, only the first 255 chars are returned
        # with trailing whitespace removed
        outstr = outstr[:255].rstrip(b' \t\n')
        return outstr

    def _interact(self, prompt_width, is_input=False):
        """Manage the interactive mode."""
        # force cursor visibility in all case
        self._cursor.set_override(True)
        self._io_streams.flush()
        try:
            # this is where we started
            start_row = self._text_screen.current_row
            furthest_left = 1 + prompt_width
            # this is where we arrow-keyed on the start line
            furthest_right = self._text_screen.current_col
            while True:
                # get one e-ASCII or dbcs code
                d = self._keyboard.get_fullchar_block()
                if not d:
                    # input stream closed
                    raise error.Exit()
                if d in (
                        ea.UP, ea.CTRL_6, ea.DOWN, ea.CTRL_MINUS,  ea.RIGHT, ea.CTRL_BACKSLASH,
                        ea.LEFT, ea.CTRL_RIGHTBRACKET, ea.HOME, ea.CTRL_k, ea.END, ea.CTRL_n
                    ):
                    # arrow keys drop us out of insert mode
                    self._set_overwrite_mode(True)
                if d == ea.CTRL_c:
                    # CTRL-C -- only caught here, not in wait_char like <CTRL+BREAK>
                    raise error.Break()
                elif d == b'\r':
                    # ENTER, CTRL+M
                    break
                elif d == b'\a':
                    # BEL, CTRL+G
                    self._sound.beep()
                elif d == b'\b':
                    # BACKSPACE, CTRL+H
                    self._text_screen.backspace(start_row, furthest_left)
                elif d == b'\t':
                    # TAB, CTRL+I
                    self._text_screen.tab(self._overwrite_mode)
                elif d == b'\n':
                    # CTRL+ENTER, CTRL+J
                    self._text_screen.line_feed()
                elif d == ea.ESCAPE:
                    # ESC, CTRL+[
                    logic_start = self._text_screen.find_start_of_line(self._text_screen.current_row)
                    if logic_start == start_row:
                        self._text_screen.clear_line(
                            logic_start, furthest_left, quirky_scrolling=is_input
                        )
                    else:
                        self._text_screen.clear_line(logic_start, 1)
                elif d in (ea.CTRL_END, ea.CTRL_e):
                    self._text_screen.clear_line(
                        self._text_screen.current_row, self._text_screen.current_col
                    )
                elif d in (ea.UP, ea.CTRL_6):
                    self._text_screen.up()
                elif d in (ea.DOWN, ea.CTRL_MINUS):
                    self._text_screen.down()
                elif d in (ea.RIGHT, ea.CTRL_BACKSLASH):
                    self._text_screen.incr_pos()
                elif d in (ea.LEFT, ea.CTRL_RIGHTBRACKET):
                    self._text_screen.decr_pos()
                elif d in (ea.CTRL_RIGHT, ea.CTRL_f):
                    self._text_screen.skip_word_right()
                elif d in (ea.CTRL_LEFT, ea.CTRL_b):
                    self._text_screen.skip_word_left()
                elif d in (ea.INSERT, ea.CTRL_r):
                    self._set_overwrite_mode(not self._overwrite_mode)
                elif d in (ea.DELETE, ea.CTRL_BACKSPACE):
                    self._text_screen.delete_fullchar()
                elif d in (ea.HOME, ea.CTRL_k):
                    self._text_screen.set_pos(1, 1)
                elif d in (ea.END, ea.CTRL_n):
                    self._text_screen.move_to_end()
                elif d in (ea.CTRL_HOME, ea.CTRL_l):
                    self._text_screen.clear_view()
                elif d == ea.CTRL_PRINT:
                    # ctrl+printscreen toggles printer copy
                    self._io_streams.toggle_output_stream(self._lpt1_file)
                else:
                    try:
                        # these are done on a less deep level than the fn key macros
                        letters = list(iterchar(ALT_KEY_REPLACE[d])) + [b' ']
                    except KeyError:
                        letters = [d]
                    for d in letters:
                        # ignore eascii by this point, but not dbcs
                        if d[:1] not in (b'\0', b'\r'):
                            if not self._overwrite_mode:
                                self._text_screen.insert_fullchars(d)
                            else:
                                # put all dbcs in before messing with cursor position
                                self._text_screen.write_chars(d, do_scroll_down=True)
                if self._text_screen.current_row == start_row:
                    furthest_left = min(self._text_screen.current_col, furthest_left)
                    furthest_right = max(self._text_screen.current_col, furthest_right)
                    if (
                            self._text_screen.current_col == self._text_screen.mode.width
                            and self._text_screen.overflow
                        ):
                        furthest_right += 1
        finally:
            self._set_overwrite_mode(True)
            # reset cursor visibility
            self._cursor.set_override(False)
        return start_row, furthest_left, furthest_right

    def _set_overwrite_mode(self, new_overwrite):
        """Set or unset the overwrite mode (INS)."""
        if new_overwrite != self._overwrite_mode:
            self._overwrite_mode = new_overwrite
            self._cursor.set_default_shape(new_overwrite)

    ##########################################################################
    # output

    def write(self, s, do_echo=True):
        """Write a string to the screen at the current position."""
        if not s:
            # don't disconnect line wrap if no output
            return
        if do_echo:
            # CR -> CRLF, CRLF -> CRLF LF
            self._io_streams.write(b''.join([(b'\r\n' if c == b'\r' else c) for c in iterchar(s)]))
        last = b''
        # if our line wrapped at the end before, it doesn't anymore
        self._text_screen.set_wrap(self.current_row, False)
        out_chars = []
        for c in iterchar(s):
            if c in b'\t\n\r\a\x0B\x0C\x1C\x1D\x1E\x1F':
                # non-printing or position-dependent chars, dump buffer first
                self._text_screen.write_chars(b''.join(out_chars), do_scroll_down=False)
                out_chars = []
                row, col = self.current_row, self.current_col
                if c == b'\t':
                    # TAB
                    num = (8 - (col - 1 - 8 * int((col-1) // 8)))
                    self._text_screen.write_chars(b' ' * num, do_scroll_down=False)
                elif c == b'\n' or c == b'\r':
                    # CR or LF
                    # note that a PRINTed LF chr$(10) does not cause a wrapped/connected line
                    # in contrast to a typed Ctrl+J
                    self._text_screen.newline(wrap=False)
                elif c == b'\a':
                    # BEL
                    self._sound.beep()
                elif c == b'\x0B':
                    # HOME
                    self._text_screen.set_pos(1, 1, scroll_ok=False)
                elif c == b'\x0C':
                    # CLS
                    self._text_screen.clear_view()
                elif c == b'\x1C':
                    # RIGHT
                    self._text_screen.set_pos(row, col + 1, scroll_ok=False)
                elif c == b'\x1D':
                    # LEFT
                    self._text_screen.set_pos(row, col - 1, scroll_ok=False)
                elif c == b'\x1E':
                    # UP
                    self._text_screen.set_pos(row - 1, col, scroll_ok=False)
                elif c == b'\x1F':
                    # DOWN
                    self._text_screen.set_pos(row + 1, col, scroll_ok=False)
            else:
                # includes \b, \0, and non-control chars
                out_chars.append(c)
            last = c
        self._text_screen.write_chars(b''.join(out_chars), do_scroll_down=False)

    def write_line(self, s=b'', do_echo=True):
        """Write a string to the screen and end with a newline."""
        self.write(b'%s\r' % (s,), do_echo)

    def list_line(self, line, newline, set_text_position=None):
        """Print a line from a program listing or EDIT prompt."""
        # no wrap if 80-column line, clear row before printing.
        # replace LF CR with LF
        line = line.replace(b'\n\r', b'\n')
        cuts = line.split(b'\n')
        for i, l in enumerate(cuts):
            if i > 0:
                # echo
                self._io_streams.write(b'\n')
                # when using LIST, we *do* print LF as a wrap
                self._text_screen.newline(wrap=True)
            self._text_screen.clear_line(self._text_screen.current_row, 1)
            self.write(l)
        if newline:
            self.write_line()
        # remove wrap after 80-column program line
        if len(line) == self.width and self._text_screen.current_row > 2:
            self._text_screen.set_wrap(self._text_screen.current_row-2, False)
        if set_text_position is not None:
            # adjust column position if logical line extends across multiple rows
            pos_row, pos_col = self._text_screen.find_position_in_line(
                self.current_row, set_text_position
            )
            self.set_pos(pos_row, pos_col)

    def start_line(self):
        """
        Move the cursor to the start of the next line, this line if empty.
        Used for prompt and error or break messages.
        """
        if self.current_col != 1:
            self._io_streams.write(b'\r\n')
            self._text_screen.set_pos(self._text_screen.current_row + 1, 1)
        # ensure line above doesn't wrap
        self._text_screen.set_wrap(self._text_screen.current_row-1, False)


    ##########################################################################
    # function key macros

    def key_(self, args):
        """KEY: show/hide/list macros."""
        command, = args
        if command == tk.ON:
            self._text_screen.show_bar(True)
        elif command == tk.OFF:
            self._text_screen.show_bar(False)
        elif command == tk.LIST:
            self._list_macros()

    def set_macro(self, num, macro):
        """Set macro for given function key."""
        # keys 30, 31 map to F11, F12 (which may or may not be definable)
        # except on Tandy where this is keys 11, 12
        if not self._tandy_fn_keys:
            if num >= 30:
                num -= 19
            elif num > 10:
                raise ValueError('Function key number out of range')
        if num > self._num_fn_keys:
            raise ValueError('Function key number out of range')
        # NUL terminates macro string, rest is ignored
        # macro starting with NUL is empty macro
        self._keyboard.set_macro(num, macro)
        self._update_bar()
        self._text_screen.redraw_bar()

    def _list_macros(self):
        """Print a list of the function key macros."""
        for i in range(self._num_fn_keys):
            text = self._keyboard.get_macro(i)
            text = b''.join(FKEY_MACRO_REPLACE_CHARS.get(s, s) for s in iterchar(text))
            self.write_line(b'F%d %s' % (i+1, text))

    def _update_bar(self):
        """Show/hide the function keys line on the active page."""
        macros = (
            self._keyboard.get_macro(_i)
            for _i in range(10)
        )
        descriptions = [
            b''.join(FKEY_MACRO_REPLACE_CHARS.get(_s, _s) for _s in iterchar(_macro[:6]))
            for _macro in macros
        ]
        self._text_screen.update_bar(descriptions)



"""
win_subprocess.py
by Valentin LAB

https://gist.github.com/vaab/2ad7051fc193167f15f85ef573e54eb9
https://bugs.python.org/issue19264
"""

# pylint: disable=no-name-in-module, import-error, no-member

import os
import ctypes
import subprocess
import _subprocess
from ctypes import byref, windll, c_wchar_p, c_void_p, Structure, sizeof, c_wchar, WinError
from ctypes.wintypes import BYTE, WORD, BOOL, DWORD, LPWSTR, LPVOID, HANDLE, LPCWSTR

##
## Types
##

CREATE_UNICODE_ENVIRONMENT = 0x00000400
LPSECURITY_ATTRIBUTES = c_void_p
LPBYTE  = ctypes.POINTER(BYTE)

class STARTUPINFOW(Structure):
    _fields_ = [
        ("cb",              DWORD),  ("lpReserved",    LPWSTR),
        ("lpDesktop",       LPWSTR), ("lpTitle",       LPWSTR),
        ("dwX",             DWORD),  ("dwY",           DWORD),
        ("dwXSize",         DWORD),  ("dwYSize",       DWORD),
        ("dwXCountChars",   DWORD),  ("dwYCountChars", DWORD),
        ("dwFillAtrribute", DWORD),  ("dwFlags",       DWORD),
        ("wShowWindow",     WORD),   ("cbReserved2",   WORD),
        ("lpReserved2",     LPBYTE), ("hStdInput",     HANDLE),
        ("hStdOutput",      HANDLE), ("hStdError",     HANDLE),
    ]

LPSTARTUPINFOW = ctypes.POINTER(STARTUPINFOW)

class PROCESS_INFORMATION(Structure):
    _fields_ = [
        ("hProcess",         HANDLE), ("hThread",          HANDLE),
        ("dwProcessId",      DWORD),  ("dwThreadId",       DWORD),
    ]

LPPROCESS_INFORMATION = ctypes.POINTER(PROCESS_INFORMATION)


class DUMMY_HANDLE(ctypes.c_void_p):

    def __init__(self, *a, **kw):
        super(DUMMY_HANDLE, self).__init__(*a, **kw)
        self.closed = False

    def Close(self):
        if not self.closed:
            windll.kernel32.CloseHandle(self)
            self.closed = True

    def __int__(self):
        return self.value


CreateProcessW = windll.kernel32.CreateProcessW
CreateProcessW.argtypes = [
    LPCWSTR, LPWSTR, LPSECURITY_ATTRIBUTES,
    LPSECURITY_ATTRIBUTES, BOOL, DWORD, LPVOID, LPCWSTR,
    LPSTARTUPINFOW, LPPROCESS_INFORMATION,
]
CreateProcessW.restype = BOOL


##
## Patched functions/classes
##

def CreateProcess(executable, args, _p_attr, _t_attr,
                  inherit_handles, creation_flags, env, cwd,
                  startup_info):
    """Create a process supporting unicode executable and args for win32

    Python implementation of CreateProcess using CreateProcessW for Win32

    """

    si = STARTUPINFOW(
        dwFlags=startup_info.dwFlags,
        wShowWindow=startup_info.wShowWindow,
        cb=sizeof(STARTUPINFOW),
        ## XXXvlab: not sure of the casting here to ints.
        hStdInput=int(startup_info.hStdInput),
        hStdOutput=int(startup_info.hStdOutput),
        hStdError=int(startup_info.hStdError),
    )

    wenv = None
    if env is not None:
        ## LPCWSTR seems to be c_wchar_p, so let's say CWSTR is c_wchar
        env = (u''.join([u'%s=%s\0' % (k, v) for k, v in env.items()])) + u'\0'
        wenv = (c_wchar * len(env))()
        wenv.value = env

    pi = PROCESS_INFORMATION()
    creation_flags |= CREATE_UNICODE_ENVIRONMENT

    if CreateProcessW(executable, args, None, None,
                      inherit_handles, creation_flags,
                      wenv, cwd, byref(si), byref(pi)):
        return (DUMMY_HANDLE(pi.hProcess), DUMMY_HANDLE(pi.hThread),
                pi.dwProcessId, pi.dwThreadId)
    raise WinError()


class Popen(subprocess.Popen):
    """This superseeds Popen and corrects a bug in cPython 2.7 implem"""

    def _execute_child(self, args, executable, preexec_fn, close_fds,
                       cwd, env, universal_newlines,
                       startupinfo, creationflags, shell, to_close,
                       p2cread, p2cwrite,
                       c2pread, c2pwrite,
                       errread, errwrite):
        """Code from part of _execute_child from Python 2.7 (9fbb65e)

        There are only 2 little changes concerning the construction of
        the the final string in shell mode: we preempt the creation of
        the command string when shell is True, because original function
        will try to encode unicode args which we want to avoid to be able to
        sending it as-is to ``CreateProcess``.

        """
        if not isinstance(args, subprocess.types.StringTypes):
            args = subprocess.list2cmdline(args)

        if startupinfo is None:
            startupinfo = subprocess.STARTUPINFO()
        if shell:
            startupinfo.dwFlags |= _subprocess.STARTF_USESHOWWINDOW
            startupinfo.wShowWindow = _subprocess.SW_HIDE
            comspec = os.environ.get("COMSPEC", u'cmd.exe')
            args = u'{} /c "{}"'.format(comspec, args)
            if (_subprocess.GetVersion() >= 0x80000000 or
                    os.path.basename(comspec).lower() == 'command.com'):
                w9xpopen = self._find_w9xpopen()
                args = u'"%s" %s' % (w9xpopen, args)
                creationflags |= _subprocess.CREATE_NEW_CONSOLE

        super(Popen, self)._execute_child(args, executable,
            preexec_fn, close_fds, cwd, env, universal_newlines,
            startupinfo, creationflags, False, to_close, p2cread,
            p2cwrite, c2pread, c2pwrite, errread, errwrite)

_subprocess.CreateProcess = CreateProcess



"""
PC-BASIC - eascii.py
Keyboard e-ASCII codes

(c) 2015--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ...compat import iteritems, unichr, SimpleNamespace


# based on Tandy-1000 basic manual, modified for IBM PC keyboard
# we don't specify the standard ASCII character values here
# nor the CTRL+a -> b'\x01' etc. series
# except where convenient

as_bytes = SimpleNamespace(

    NUL = b'\0\0',

    CTRL_b = b'\x02',
    CTRL_c = b'\x03',
    CTRL_d = b'\x04',
    CTRL_e = b'\x05',
    CTRL_f = b'\x06',
    CTRL_k = b'\x0B',
    CTRL_l = b'\x0C',
    CTRL_n = b'\x0E',
    CTRL_r = b'\x12',
    CTRL_z = b'\x1A',

    ESCAPE = b'\x1B',
    SHIFT_ESCAPE = b'\x1B',
    CTRL_ESCAPE = b'\x1B',

    BACKSPACE = b'\x08',
    SHIFT_BACKSPACE = b'\x08',
    CTRL_BACKSPACE = b'\x7F',
    ALT_BACKSPACE = b'\0\x8C',

    TAB = b'\x09',
    SHIFT_TAB = b'\0\x0F',
    CTRL_TAB = b'\0\x8D',
    ALT_TAB = b'\0\x8E',

    RETURN = b'\r',
    SHIFT_RETURN = b'\r',
    CTRL_RETURN = b'\n',
    ALT_RETURN = b'\0\x8F',

    SPACE = b' ',
    SHIFT_SPACE = b' ',
    CTRL_SPACE = b' ',
    ALT_SPACE = b'\0 ',

    CTRL_PRINT = b'\0\x72',
    ALT_PRINT = b'\0\x46',

    INSERT = b'\0\x52',
    SHIFT_INSERT = b'\0\x52',

    DELETE = b'\0\x53',
    SHIFT_DELETE = b'\0\x53',

    CTRL_2 = b'\0\x03',
    CTRL_6 = b'\x1E',
    CTRL_MINUS = b'\x1F',
    SHIFT_KP5 = b'5',
    ALT_KP5 = b'\x05',

    CTRL_BACKSLASH = b'\x1C',
    # CTRL+]
    CTRL_RIGHTBRACKET = b'\x1D',

    # Alt codes

    ALT_1 = b'\0\x78',
    ALT_2 = b'\0\x79',
    ALT_3 = b'\0\x7A',
    ALT_4 = b'\0\x7B',
    ALT_5 = b'\0\x7C',
    ALT_6 = b'\0\x7D',
    ALT_7 = b'\0\x7E',
    ALT_8 = b'\0\x7F',
    ALT_9 = b'\0\x80',
    ALT_0 = b'\0\x81',
    ALT_MINUS = b'\0\x82',
    ALT_EQUALS = b'\0\x83',

    ALT_q = b'\0\x10',
    ALT_w = b'\0\x11',
    ALT_e = b'\0\x12',
    ALT_r = b'\0\x13',
    ALT_t = b'\0\x14',
    ALT_y = b'\0\x15',
    ALT_u = b'\0\x16',
    ALT_i = b'\0\x17',
    ALT_o = b'\0\x18',
    ALT_p = b'\0\x19',

    ALT_a = b'\0\x1E',
    ALT_s = b'\0\x1F',
    ALT_d = b'\0\x20',
    ALT_f = b'\0\x21',
    ALT_g = b'\0\x22',
    ALT_h = b'\0\x23',
    ALT_j = b'\0\x24',
    ALT_k = b'\0\x25',
    ALT_l = b'\0\x26',

    ALT_z = b'\0\x2C',
    ALT_x = b'\0\x2D',
    ALT_c = b'\0\x2E',
    ALT_v = b'\0\x2F',
    ALT_b = b'\0\x30',
    ALT_n = b'\0\x31',
    ALT_m = b'\0\x32',

    # function keys

    F1 = b'\0\x3B',
    F2 = b'\0\x3C',
    F3 = b'\0\x3D',
    F4 = b'\0\x3E',
    F5 = b'\0\x3F',
    F6 = b'\0\x40',
    F7 = b'\0\x41',
    F8 = b'\0\x42',
    F9 = b'\0\x43',
    F10 = b'\0\x44',

    SHIFT_F1 = b'\0\x54',
    SHIFT_F2 = b'\0\x55',
    SHIFT_F3 = b'\0\x56',
    SHIFT_F4 = b'\0\x57',
    SHIFT_F5 = b'\0\x58',
    SHIFT_F6 = b'\0\x59',
    SHIFT_F7 = b'\0\x5A',
    SHIFT_F8 = b'\0\x5B',
    SHIFT_F9 = b'\0\x5C',
    SHIFT_F10 = b'\0\x5D',

    CTRL_F1 = b'\0\x5E',
    CTRL_F2 = b'\0\x5F',
    CTRL_F3 = b'\0\x60',
    CTRL_F4 = b'\0\x61',
    CTRL_F5 = b'\0\x62',
    CTRL_F6 = b'\0\x63',
    CTRL_F7 = b'\0\x64',
    CTRL_F8 = b'\0\x65',
    CTRL_F9 = b'\0\x66',
    CTRL_F10 = b'\0\x67',

    ALT_F1 = b'\0\x68',
    ALT_F2 = b'\0\x69',
    ALT_F3 = b'\0\x6A',
    ALT_F4 = b'\0\x6B',
    ALT_F5 = b'\0\x6C',
    ALT_F6 = b'\0\x6D',
    ALT_F7 = b'\0\x6E',
    ALT_F8 = b'\0\x6F',
    ALT_F9 = b'\0\x70',
    ALT_F10 = b'\0\x71',

    # numeric keypad
    HOME = b'\0\x47',
    UP = b'\0\x48',
    PAGEUP = b'\0\x49',
    LEFT = b'\0\x4B',
    RIGHT = b'\0\x4D',
    END = b'\0\x4F',
    DOWN = b'\0\x50',
    PAGEDOWN = b'\0\x51',

    SHIFT_HOME = b'\0\x47',
    SHIFT_UP = b'\0\x48',
    SHIFT_PAGEUP = b'\0\x49',
    SHIFT_LEFT = b'\0\x87',
    SHIFT_RIGHT = b'\0\x88',
    SHIFT_END = b'\0\x4F',
    SHIFT_DOWN = b'\0\x50',
    SHIFT_PAGEDOWN = b'\0\x51',

    CTRL_HOME = b'\0\x77',
    CTRL_PAGEUP = b'\0\x84',
    CTRL_LEFT = b'\0\x73',
    CTRL_RIGHT = b'\0\x74',
    CTRL_END = b'\0\x75',
    CTRL_PAGEDOWN = b'\0\x76',

    # Tandy e-ASCII codes
    F11 = b'\0\x98',
    F12 = b'\0\x99',
    SHIFT_F11 = b'\0\xA2',
    SHIFT_F12 = b'\0\xA3',
    CTRL_F11 = b'\0\xAC',
    CTRL_F12 = b'\0\xAD',
    ALT_F11 = b'\0\xB6',
    ALT_F12 = b'\0\xB7',
)

as_unicode = SimpleNamespace(**{
    key: value.decode('latin-1')
    for key, value in iteritems(as_bytes.__dict__)
})



"""
PC-BASIC - config.py
Configuration file and command-line options parser

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import io
import sys
import logging
import zipfile
import locale
import shutil
import codecs
from collections import deque

from .compat import iteritems, text_type, iterchar
from .compat import configparser
from .compat import WIN32, get_short_pathname, argv, getcwdu
from .compat import USER_CONFIG_HOME, USER_DATA_HOME, PY2
from .compat import split_quoted, split_pair
from .compat import console, IS_CONSOLE_APP, stdio
from .compat import TemporaryDirectory

from .data import CODEPAGES, FONTS, PROGRAMS, ICON
from .basic import VERSION, NAME
from . import data


# base directory name
MAJOR_VERSION = u'.'.join(VERSION.split(u'.')[:2])
BASENAME = u'pcbasic-{0}'.format(MAJOR_VERSION)

# user configuration and state directories
USER_CONFIG_DIR = os.path.join(USER_CONFIG_HOME, BASENAME)
STATE_PATH = os.path.join(USER_DATA_HOME, BASENAME)

# default config file name
CONFIG_NAME = u'PCBASIC.INI'

# user and local config files
USER_CONFIG_PATH = os.path.join(USER_CONFIG_DIR, CONFIG_NAME)

# save-state file name
STATE_NAME = 'pcbasic.session'

# @: target drive for bundled programs
PROGRAM_PATH = os.path.join(STATE_PATH, u'bundled_programs')

# maximum memory size
MAX_MEMORY_SIZE = 65534

# format for log files
LOGGING_FORMAT = u'[%(asctime)s.%(msecs)04d] %(levelname)s: %(message)s'
LOGGING_FORMATTER = logging.Formatter(fmt=LOGGING_FORMAT, datefmt=u'%H:%M:%S')

# drive letters except @, bytes constant
UPPERCASE = b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

# bool strings
TRUES = (u'YES', u'TRUE', u'ON', u'1')
FALSES = (u'NO', u'FALSE', u'OFF', u'0')


##############################################################################
# version checks

# minimum required python versions
MIN_PYTHON2_VERSION = (2, 7, 12)
MIN_PYTHON3_VERSION = (3, 5, 0)

def _validate_version():
    """Initial validations."""
    # sys.version_info tuple's first three elements are guaranteed to be ints
    python_version = sys.version_info[:3]
    if (
            (python_version[0] == 2 and python_version < MIN_PYTHON2_VERSION) or
            (python_version[0] == 3 and python_version < MIN_PYTHON3_VERSION)
        ):
        msg = (
            'PC-BASIC requires Python version %d.%d.%d, ' % MIN_PYTHON2_VERSION +
            'version %d.%d.%d, or higher. ' % MIN_PYTHON3_VERSION +
            'You have %d.%d.%d.' % python_version
        )
        logging.fatal(msg)
        raise ImportError(msg)

# raise ImportError if incorrect Python version
_validate_version()


##############################################################################
# Default preset definitions
# For example, --preset=tandy will load all options in the [tandy] section.
# Preset options override default options.
# Presets can also be added by adding a section in brackets in the user configuration file
# e.g., a section headed [myconfig] will be loaded with --preset=myconfig

PRESETS = {
    u'strict': {
        u'hide-listing': u'65530',
        u'text-encoding': u'',
        u'soft-linefeed': u'False',
        u'hide-protected': u'True',
        u'allow-code-poke': u'True',
        u'prevent-close': u'True',
        u'ctrl-c-break': u'False',
    },
    u'gwbasic': {
        u'syntax': u'gwbasic',
    },
    u'basica': {
        u'reserved-memory': u'789',
    },
    u'pcjr': {
        u'syntax': u'pcjr',
        u'term': os.path.join(PROGRAM_PATH, 'PCTERM.BAS'),
        u'video': u'pcjr',
        u'font': u'vga',
        u'codepage': u'437',
        u'reserved-memory': u'4035',
        u'text-width': u'40',
        u'video-memory': u'16384',
    },
    u'tandy': {
        u'syntax': u'tandy',
        u'video': u'tandy',
        u'font': u'tandy2',
        u'codepage': u'437',
        u'aspect': u'3072,2000',
        u'max-reclen': u'255',
        u'reserved-memory': u'3240',
        u'video-memory': u'16384',
    },
    u'cga': {
        u'video': u'cga',
        u'font': u'cga',
        u'codepage': u'437',
        u'text-width': u'40',
    },
    u'ega': {
        u'video': u'ega',
        u'font': u'vga',
    },
    u'mda': {
        u'video': u'mda',
        u'font': u'cga,mda',
        u'codepage': u'437',
        u'monitor': u'mono',
    },
    u'hercules': {
        u'video': u'hercules',
        u'font': u'cga,mda',
        u'codepage': u'437',
        u'monitor': u'mono',
    },
    u'olivetti': {
        u'video': u'olivetti',
        u'font': u'cga,olivetti',
        u'codepage': u'437',
    },
    u'vga': {
        u'video': u'vga',
        u'font': u'vga',
        u'codepage': u'437',
    },
}

# by default, load what's in section [pcbasic] and override with anything
DEFAULT_SECTION = [u'pcbasic']


##############################################################################
# short-form arguments

SHORT_ARGS = {
    u'd': (u'double', u'True'),
    u'h': (u'help', u'True'),
    u'q': (u'quit', u'True'),
    u'v': (u'version', u'True'),
    u'w': (u'wait', u'True'),
    u'b': (u'interface', u'cli'),
    u't': (u'interface', u'text'),
    u'n': (u'interface', u'none'),
    u'f': (u'max-files', None),
    u's': (u'max-reclen', None),
    u'l': (u'load', None),
    u'r': (u'run', None),
    u'e': (u'exec', None),
    u'k': (u'keys', None),
    u'i': (u'input', None),
    u'o': (u'output', None),
}

# -c means the same as -q -n -e
SHORTHAND = {
    u'c': u'qne',
}

##############################################################################
# GWBASIC-style options
# GWBASIC [prog] [<inp] [[>]>outp] [/f:n] [/i] [/s:n] [/c:n] [/m:[n][,m]] [/d]
#   /d      Allow double-precision ATN, COS, EXP, LOG, SIN, SQR, and TAN.
#   /f:n    Set maximum number of open files to n. Default is 3.
#           Each additional file reduces free memory by 322 bytes.
#   /s:n    Set the maximum record length for RANDOM files.
#           Default is 128, maximum is 32768.
#   /c:n    Set the COM receive buffer to n bytes.
#           If n==0, disable the COM ports.
#   /i      Statically allocate file control blocks and data buffer.
#           NOTE: this appears to be always the case in GW-BASIC, as here.
#   /m:n,m  Set the highest memory location to n (default 65534) and maximum
#           BASIC memory to m*16 bytes (default is all available).

GW_OPTIONS = {
    u'<': 'input',
    u'>': 'output',
    u'>>': 'output:append',
    u'/d': 'double',
    u'/i': '',
    u'/f': 'max-files',
    u'/s': 'max-reclen',
    u'/c': 'serial-buffer-size',
    u'/m': 'max-memory',
}


##############################################################################
# long-form arguments

def _check_text_encoding(arg):
    """Check if text-encoding argument is acceptable."""
    try:
        codecs.lookup(arg)
    except LookupError:
        return False
    return True

def _check_max_memory(arglist):
    """Check if max-memory argument is acceptable."""
    mem_sizes = [arglist[0], arglist[1]*16 if arglist[1] else None]
    if min((mem_size for mem_size in mem_sizes if mem_size), default=MAX_MEMORY_SIZE) > MAX_MEMORY_SIZE:
        logging.warning(u'max-memory value > %s', MAX_MEMORY_SIZE)
        return False
    return True


# number of positional arguments
NUM_POSITIONAL = 2

ARGUMENTS = {
    u'input': {u'type': u'string', u'default': u'', },
    u'output': {u'type': u'string', u'default': u'', },
    u'interface': {
        u'type': u'string', u'default': u'',
        u'choices': (
            u'', u'none', u'cli', u'text', u'graphical',
            u'ansi', u'curses', u'pygame', u'sdl2'
        ),
    },
    u'sound': {
        u'type': u'string', u'default': u'true',
        u'choices': (u'true', u'false', u'none', u'beep', u'portaudio', u'sdl2', u'interface'),
    },
    u'load': {u'type': u'string', u'default': u'', },
    u'run': {u'type': u'string', u'default': u'',  },
    u'convert': {u'type': u'string', u'default': u'', },
    u'help': {u'type': u'bool', u'default': False, },
    u'keys': {u'type': u'string', u'default': u'', },
    u'exec': {u'type': u'string', u'default': u'', },
    u'quit': {u'type': u'bool', u'default': False,},
    u'double': {u'type': u'bool', u'default': False,},
    u'max-files': {u'type': u'int', u'default': 3,},
    u'max-reclen': {u'type': u'int', u'default': 128,},
    u'serial-buffer-size': {u'type': u'int', u'default': 256,},
    u'peek': {u'type': u'string', u'list': u'*', u'default': [],},
    u'lpt1': {u'type': u'string', u'default': u'PRINTER:',},
    u'lpt2': {u'type': u'string', u'default': u'',},
    u'lpt3': {u'type': u'string', u'default': u'',},
    u'cas1': {u'type': u'string', u'default': u'',},
    u'com1': {u'type': u'string', u'default': u'',},
    u'com2': {u'type': u'string', u'default': u'',},
    u'codepage': {u'type': u'string', u'choices': CODEPAGES, u'default': u'437',},
    u'font': {
        u'type': u'string', u'list': u'*', u'choices': FONTS,
        u'default': [u'default'],
    },
    u'dimensions': {u'type': u'int', u'list': 2, u'default': [],},
    u'fullscreen': {u'type': u'bool', u'default': False,},
    u'prevent-close': {u'type': u'bool', u'default': False,},
    u'debug': {u'type': u'bool', u'default': False,},
    u'hide-listing': {u'type': u'int', u'default': 65535,},
    u'hide-protected': {u'type': u'bool', u'default': False,},
    u'mount': {u'type': u'string', u'list': u'*', u'default': [],},
    u'resume': {u'type': u'bool', u'default': False,},
    u'syntax': {
        u'type': u'string', u'choices': (u'advanced', u'gwbasic', u'pcjr', u'tandy'),
        u'default': u'advanced',
    },
    u'term': {u'type': u'string', u'default': u'',},
    u'video': {
        u'type': u'string', u'default': 'vga',
        u'choices': (
            u'vga', u'ega', u'cga', u'mda',
            u'pcjr', u'tandy', u'hercules', u'olivetti'
        ),
    },
    u'text-encoding': {u'type': u'string', u'default': u'', u'check': _check_text_encoding},
    u'soft-linefeed': {u'type': u'bool', u'default': False,},
    u'border': {u'type': u'int', u'default': 5,},
    u'mouse-clipboard': {u'type': u'bool', u'default': True,},
    u'state': {u'type': u'string', u'default': u'',},
    u'monitor': {
        u'type': u'string',
        u'choices': (u'rgb', u'composite', u'green', u'amber', u'grey', u'mono'),
        u'default': u'rgb',
    },
    u'aspect': {u'type': u'int', u'list': 2, u'default': [4, 3],},
    u'scaling': {
        u'type': u'string', u'choices':(u'smooth', u'native', u'crisp'),
        u'default': u'smooth',
    },
    u'version': {u'type': u'bool', u'default': False,},
    u'config': {u'type': u'string', u'default': u'',},
    u'logfile': {u'type': u'string', u'default': u'',},
    # negative list length means 'optionally up to'
    u'max-memory': {u'type': u'int', u'list': -2, u'default': [MAX_MEMORY_SIZE, 4096], u'listcheck': _check_max_memory},
    u'allow-code-poke': {u'type': u'bool', u'default': False,},
    u'reserved-memory': {u'type': u'int', u'default': 3429,},
    u'caption': {u'type': u'string', u'default': NAME,},
    u'text-width': {u'type': u'int', u'choices':(u'40', u'80'), u'default': 80,},
    u'video-memory': {u'type': u'int', u'default': 262144,},
    u'shell': {u'type': u'string', u'default': u'',},
    u'ctrl-c-break': {u'type': u'bool', u'default': True,},
    u'wait': {u'type': u'bool', u'default': False,},
    u'current-device': {u'type': u'string', u'default': ''},
    u'extension': {u'type': u'string', u'list': u'*', u'default': []},
    u'options': {u'type': u'string', u'default': ''},
    # depecated argument, use text-encoding instead
    u'utf8': {u'type': u'bool', u'default': False,},
}


##########################################################################
# logging

class Lumberjack(object):
    """Logging manager."""

    def __init__(self):
        """Set up the global logger temporarily until we know the log stream."""
        # include messages from warnings madule in the logs
        logging.captureWarnings(True)
        # we use the awkward logging interface as we can only use basicConfig once
        # get the root logger
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.INFO)
        # send to a buffer until we know where to log to
        self._logstream = io.StringIO()
        handler = logging.StreamHandler(self._logstream)
        handler.setFormatter(LOGGING_FORMATTER)
        root_logger.addHandler(handler)

    def reset(self):
        """Reset root logger."""
        # o dear o dear what a horrible API
        root_logger = logging.getLogger()
        # remove all old handlers: temporary ones we set as well as any default ones
        for handler in root_logger.handlers:
            root_logger.removeHandler(handler)
        return root_logger

    def prepare(self, logfile, debug):
        """Set up the global logger."""
        # get log stream and level from options
        loglevel = logging.DEBUG if debug else logging.INFO
        root_logger = self.reset()
        root_logger.setLevel(loglevel)
        if logfile:
            logstream = io.open(logfile, 'w', encoding='utf_8', errors='replace')
        else:
            logstream = stdio.stderr
        # write out cached logs
        logstream.write(self._logstream.getvalue())
        handler = logging.StreamHandler(logstream)
        handler.setFormatter(LOGGING_FORMATTER)
        root_logger.addHandler(handler)


##########################################################################
# write default config file

CONFIG_HEADER = (
    u"# PC-BASIC configuration file.\n"
    u"# Edit this file to change your default settings or add presets.\n"
    u"# Changes to this file will not affect any other users of your computer.\n"
    u"# All lines starting with # are comments and have no effect.\n"
    u"# Thus, to use one of the example options below, "
    u"you need to remove the # at the start of the line.\n"
    u"\n"
    u"[pcbasic]\n"
    u"# Use the [pcbasic] section to specify options you want to be enabled by default.\n"
    u"# See the documentation or run pcbasic -h for a list of available options.\n"
    u"# for example (for version '%s'):\n" % VERSION
)

CONFIG_FOOTER = (
    u"\n\n# To add presets, create a section header between brackets and put the \n"
    u"# options you need below it, like this:\n"
    u"# [your_preset]\n"
    u"# border=0\n"
    u"# \n"
    u"# You will then be able to load these options with --preset=your_preset.\n"
    u"# If you choose the same name as a system preset, PC-BASIC will use your\n"
    u"# options for that preset and not the system ones. This is not recommended.\n"
)


def _build_default_config_file(file_name):
    """Write a default config file."""
    argnames = sorted(ARGUMENTS.keys())
    try:
        with io.open(file_name, 'w', encoding='utf_8_sig', errors='replace') as f:
            f.write(CONFIG_HEADER)
            for a in argnames:
                try:
                    f.write(
                        u'## choices: %s\n' %
                        u', '.join(u'%s' % (_s,) for _s in ARGUMENTS[a][u'choices'])
                    )
                except(KeyError, TypeError):
                    pass
                try:
                    # check if it's a list
                    ARGUMENTS[a][u'list']
                    formatted = u','.join(u'%s' % (_s,) for _s in ARGUMENTS[a][u'default'])
                except(KeyError, TypeError):
                    formatted = u'%s' % (ARGUMENTS[a][u'default'],)
                f.write(u'#%s=%s\n' % (a, formatted))
            f.write(CONFIG_FOOTER)
    except (OSError, IOError):
        # can't create file, ignore. we'll get a message later.
        pass


##############################################################################
# settings container

def _store_bundled_programs(PROGRAM_PATH):
    """Retrieve contents of BASIC programs."""
    for name in PROGRAMS:
        with io.open(os.path.join(PROGRAM_PATH, name), 'wb') as f:
            f.write(data.read_program_file(name))


class Settings(object):
    """Read and retrieve command-line settings and options."""

    def __init__(self, temp_dir, arguments):
        """Initialise settings."""
        # arguments should be unicode
        if not arguments:
            self._uargv = argv[1:]
        else:
            self._uargv = list(arguments)
        lumberjack = Lumberjack()
        try:
            # create state path if needed
            if not os.path.exists(STATE_PATH):
                os.makedirs(STATE_PATH)
            # create user config file if needed
            if not os.path.exists(USER_CONFIG_PATH):
                try:
                    os.makedirs(USER_CONFIG_DIR)
                except OSError:
                    pass
                _build_default_config_file(USER_CONFIG_PATH)
            # create @: drive if not present
            if not os.path.exists(PROGRAM_PATH):
                os.makedirs(PROGRAM_PATH)
                # unpack bundled programs
                _store_bundled_programs(PROGRAM_PATH)
            # store options in options dictionary
            self._options = ArgumentParser().retrieve_options(self._uargv, temp_dir)
        except:
            # avoid losing exception messages occuring while logging was disabled
            lumberjack.reset()
            raise
        # prepare global logger for use by main program
        lumberjack.prepare(self.get('logfile'), self.get('debug'))
        self._session_params = None

    def get(self, name, get_default=True):
        """Get value of option; choose whether to get default or None (unspecified) or '' (empty)."""
        try:
            value = self._options[name]
            if get_default and (value is None or value == u'' or value == []):
                raise KeyError
        except KeyError:
            if get_default:
                try:
                    value = ARGUMENTS[name][u'default']
                except KeyError:
                    if name in range(NUM_POSITIONAL):
                        return u''
            else:
                value = None
        return value

    ##########################################################################
    # session parameters

    @property
    def session_params(self):
        """Return a dictionary of parameters for the Session object."""
        if self._session_params:
            return self._session_params
        # don't parse any options on --resume
        # except redirects
        if self.get('resume'):
            params = self._add_implicit_redirects()
            return params
        # preset PEEK values
        peek_values = {}
        try:
            for a in self.get('peek'):
                seg, addr, val = a.split(u':')
                peek_values[int(seg)*0x10 + int(addr)] = int(val)
        except (TypeError, ValueError):
            pass
        # devices and mounts
        device_params = {
            key.upper(): self.get(key)
            for key in ('lpt1', 'lpt2', 'lpt3', 'com1', 'com2', 'cas1')
        }
        current_device, mount_dict = self._get_drives()
        device_params.update(mount_dict)
        # memory setting
        max_list = self.get('max-memory')
        max_list[1] = max_list[1]*16 if max_list[1] else max_list[0]
        max_list[0] = max_list[0] or max_list[1]
        # codepage parameters
        codepage_params = self.get('codepage').split(u':')
        codepage_dict = data.read_codepage(codepage_params[0])
        nobox = len(codepage_params) > 1 and codepage_params[1] == u'nobox'
        # video parameters
        video_params = self.get('video').split(u':')
        # redirects
        params = self._get_redirects()
        params.update({
            'syntax': self.get('syntax'),
            'video': video_params[0],
            'codepage': codepage_dict,
            'box_protect': not nobox,
            'monitor': self.get('monitor'),
            # screen settings
            'text_width': self.get('text-width'),
            'video_memory': self.get('video-memory'),
            'font': data.read_fonts(codepage_dict, self.get('font')),
            # find program for PCjr TERM command
            'term': self.get('term'),
            'shell': self.get('shell'),
            'double': self.get('double'),
            # device settings
            'devices': device_params,
            'current_device': current_device,
            'serial_buffer_size': self.get('serial-buffer-size'),
            # text file parameters
            'textfile_encoding': self.get('text-encoding'),
            'soft_linefeed': self.get('soft-linefeed'),
            # keyboard settings
            'ctrl_c_is_break': self.get('ctrl-c-break'),
            # program parameters
            'hide_listing': self.get('hide-listing'),
            'hide_protected': self.get('hide-protected'),
            'allow_code_poke': self.get('allow-code-poke'),
            'rebuild_offsets': not self.convert,
            # max available memory to BASIC (set by /m)
            'max_memory': min(max_list) or 65534,
            # maximum record length (-s)
            'max_reclen': max(1, min(32767, self.get('max-reclen'))),
            # number of file records
            'max_files': self.get('max-files'),
            # first field buffer address (workspace size; 3429 for gw-basic)
            'reserved_memory': self.get('reserved-memory'),
            'peek_values': peek_values,
            'extension': self.get('extension'),
            # ignore key buffer in console-based interfaces, to allow pasting text in console
            'check_keybuffer_full': self.get('interface') not in ('cli', 'text', 'ansi', 'curses'),
        })
        # deprecated arguments
        if self.get('utf8', get_default=False) is not None:
            if self.get('text-encoding', get_default=False) is not None:
                logging.warning(
                    'Deprecated option `utf8` ignored: `text-encoding` takes precedence.'
                )
            else:
                logging.warning('Option `utf8` is deprecated; use `text-encoding=utf-8` instead.')
                params['textfile_encoding'] = u'utf-8' if self.get('utf8') else u''
        self._session_params = params
        return params

    def _get_redirects(self):
        """Determine which i/o streams to attach based on config choices."""
        input_streams, output_streams = [], []
        # input redirects
        infile_params = self.get('input').split(u':')
        if infile_params[0].upper() in (u'STDIO', u'STDIN'):
            if u'RAW' in (_x.upper() for _x in infile_params):
                input_streams.append(stdio.stdin.buffer)
            else:
                input_streams.append(stdio.stdin)
        else:
            if len(infile_params) > 1 and infile_params[0].upper() == u'FILE':
                infile = infile_params[1]
            else:
                infile = infile_params[0]
            if infile:
                try:
                    input_streams.append(io.open(infile, 'rb'))
                except EnvironmentError as e:
                    logging.warning(u'Could not open input file `%s`: %s', infile, e.strerror)
        # output redirects
        outfile_params = self.get('output').split(u':')
        if outfile_params[0].upper() in (u'STDIO', u'STDOUT'):
            if u'RAW' in (_x.upper() for _x in outfile_params):
                output_streams.append(stdio.stdout.buffer)
            else:
                output_streams.append(stdio.stdout)
        else:
            if len(outfile_params) > 1 and outfile_params[0].upper() == u'FILE':
                outfile_params = outfile_params[1:]
            outfile = outfile_params[0]
            append = len(outfile_params) > 1 and outfile_params[1].lower() == u'append'
            if outfile:
                try:
                    output_streams.append(io.open(outfile, 'ab' if append else 'wb'))
                except EnvironmentError as e:
                    logging.warning(u'Could not open output file `%s`: %s', outfile, e.strerror)
        return self._add_implicit_redirects(input_streams, output_streams)

    def _add_implicit_redirects(self, input_streams=(), output_streams=()):
        """Determine which i/o streams to attach implicitly."""
        input_streams = list(input_streams)
        output_streams = list(output_streams)
        # add stdio if redirected or no interface
        if stdio.stdin not in input_streams and stdio.stdin.buffer not in input_streams:
            if IS_CONSOLE_APP and not stdio.stdin.isatty():
                # redirected on console; use bytes stream
                input_streams.append(stdio.stdin.buffer)
            elif IS_CONSOLE_APP and not self.interface:
                # no interface & on console; use unicode stream
                input_streams.append(stdio.stdin)
        # redirect output as well if input is redirected, but not the other way around
        # this is because (1) GW-BASIC does this from the DOS prompt
        # (2) otherwise we don't see anything - we quit after input closes
        # isatty is also false if we run as a GUI exe, so check that here
        if stdio.stdout not in output_streams and stdio.stdout.buffer not in output_streams:
            if IS_CONSOLE_APP and (not stdio.stdout.isatty() or not stdio.stdin.isatty()):
                output_streams.append(stdio.stdout.buffer)
            elif IS_CONSOLE_APP and not self.interface:
                output_streams.append(stdio.stdout)
        return {
            'output_streams': output_streams,
            'input_streams': input_streams,
        }

    def _get_drives(self):
        """Assign disk locations to disk devices."""
        # always get current device
        current_device = self.get('current-device').upper()
        # build mount dictionary
        mount_list = self.get('mount', False)
        if mount_list is None:
            mount_dict = self._get_default_drives()
            if not current_device:
                current_device = self._get_default_current_device()
        else:
            mount_dict = self._get_drives_from_list(mount_list)
        # directory for bundled BASIC programs accessible through @:
        mount_dict[b'@'] = PROGRAM_PATH
        # if Z: not specified, override it to avoid mounting through Session default
        if b'Z' not in mount_dict:
            mount_dict[b'Z'] = None
        return current_device, mount_dict

    def _get_drives_from_list(self, mount_list):
        """Assign drive letters based on mount specification."""
        mount_dict = {}
        for spec in mount_list:
            # the last one that's specified will stick
            try:
                letter, path = spec.split(u':', 1)
                try:
                    letter = letter.encode('ascii').upper()
                except UnicodeError:
                    logging.error(u'Could not mount `%s`: invalid drive letter', spec)
                # take abspath first to ensure unicode, realpath gives bytes for u'.'
                path = os.path.realpath(os.path.abspath(path))
                if not os.path.isdir(path):
                    logging.error(u'Could not mount `%s`: not a directory', spec)
                else:
                    mount_dict[letter] = path
            except (TypeError, ValueError) as e:
                logging.error(u'Could not mount `%s`: %s', spec, e)
        return mount_dict

    def _get_default_drives(self):
        """Assign default drive letters."""
        mount_dict = {}
        if WIN32:
            # get all drives in use by windows
            # if started from CMD.EXE, get the 'current working dir' for each drive
            # if not in CMD.EXE, there's only one cwd
            save_current = getcwdu()
            for letter in iterchar(UPPERCASE):
                try:
                    os.chdir(letter + b':')
                    cwd = get_short_pathname(getcwdu()) or getcwdu()
                except EnvironmentError:
                    # doesn't exist or can't access, do not mount this drive
                    pass
                else:
                    path, cwd = os.path.splitdrive(cwd)
                    if path:
                        # cwd must not start with \\
                        if cwd[:1] == u'\\':
                            path += u'\\'
                            cwd = cwd[1:]
                        if cwd:
                            mount_dict[letter] = u':'.join((path, cwd))
                        else:
                            mount_dict[letter] = path
                    else:
                        logging.warning('Not mounting `%s`: no drive letter.', cwd)
            os.chdir(save_current)
        else:
            # non-Windows systems simply have 'Z:' set to their their cwd by default
            mount_dict[b'Z'] = getcwdu()
        return mount_dict

    def _get_default_current_device(self):
        """Get the current drive letter or Z:"""
        if WIN32:
            letter, _ = os.path.splitdrive(os.path.abspath(getcwdu()))
            try:
                current_device = letter.encode('ascii')
            except UnicodeError:
                pass
        else:
            current_device = b'Z'
        return current_device

    ##########################################################################
    # interface parameters

    @property
    def interface(self):
        """Run with interface."""
        return self.get('interface') != 'none'

    @property
    def iface_params(self):
        """Dict of interface parameters."""
        interface = self.get('interface')
        # categorical interfaces
        categories = {
            'text': ('ansi', 'curses'),
            'graphical': ('sdl2', 'pygame'),
        }
        if not interface:
            # default: try graphical first, then text, then cli
            iface_list = categories['graphical'] + categories['text'] + ('cli',)
        else:
            try:
                iface_list = categories[interface]
            except KeyError:
                iface_list = (interface,)
        iface_params = {
            'try_interfaces': iface_list,
            'audio_override': self.get('sound') not in ('true', 'interface') and self.get('sound'),
        }
        iface_params.update(self._get_video_parameters())
        iface_params.update(self._get_audio_parameters())
        return iface_params

    def _get_video_parameters(self):
        """Return a dictionary of parameters for the video plugin."""
        return {
            'dimensions': self.get('dimensions'),
            'aspect_ratio': self.get('aspect'),
            'border_width': self.get('border'),
            'scaling': self.get('scaling'),
            'fullscreen': self.get('fullscreen'),
            'prevent_close': self.get('prevent-close'),
            'caption': self.get('caption'),
            'mouse_clipboard': self.get('mouse-clipboard'),
            'icon': ICON,
            'wait': self.get('wait'),
            }

    def _get_audio_parameters(self):
        """Return a dictionary of parameters for the audio plugin."""
        return {}


    ##########################################################################
    # launch parameters

    @property
    def launch_params(self):
        """Dict of launch parameters."""
        # build list of commands to execute on session startup
        commands = []
        greeting = False
        if not self.get('resume'):
            run = (
                # positional argument and --load or -l not specified (empty is specified)
                (self.get(0) and self.get('load', get_default=False) is None)
                # or run specified explicitly
                or self.get('run')
            )
            # treat colons as CRs
            commands = split_quoted(self.get('exec'), split_by=u':', quote=u'"', strip_quotes=False)
            # note that executing commands (or RUN) will suppress greeting
            # following GW, don't write greeting for redirected input or command-line filter run
            greeting = not run and not commands and not self.session_params['input_streams']
            if run:
                commands.append('RUN')
            if self.get('quit'):
                commands.append('SYSTEM')
        launch_params = {
            'prog': self.get('run') or self.get('load') or self.get(0),
            'resume': self.get('resume'),
            'greeting': greeting,
            'state_file': self._get_state_file(),
            'commands': commands,
            # inserted keystrokes
            # we first need to encode the unicode to bytes before we can decode it
            # this preserves unicode as \x (if latin-1) and \u escapes
            'keys': self.get('keys').encode('ascii', 'backslashreplace').decode('unicode-escape'),
            'debug': self.get('debug'),
            }
        launch_params.update(self.session_params)
        return launch_params

    def _get_state_file(self):
        """Name of state file"""
        state_name = self.get('state') or STATE_NAME
        if not os.path.exists(state_name):
            state_name = os.path.join(STATE_PATH, state_name)
        return state_name


    ##########################################################################
    # other calls

    @property
    def guard_params(self):
        """Dict of exception guard parameters."""
        return {
            'uargv': self._uargv,
            'log_dir': STATE_PATH,
            }

    @property
    def conv_params(self):
        """Get parameters for file conversion."""
        # conversion output
        # argument is mode
        mode = self.get('convert', get_default=False)
        # keep uppercase first letter
        mode = mode[0].upper() if mode else 'A'
        name_in = (self.get(0) or self.get('run') or self.get('load'))
        name_out = self.get(1)
        return mode, name_in, name_out

    @property
    def version(self):
        """Version operating mode."""
        return self.get('version')

    @property
    def help(self):
        """Help operating mode."""
        return self.get('help')

    @property
    def convert(self):
        """Converter operating mode."""
        return self.get('convert', get_default=False) is not None

    @property
    def debug(self):
        """Debugging mode."""
        return self.get('debug')


##############################################################################
# argument parsing

class ArgumentParser(object):
    """Parse PC-BASIC config file and command-line arguments."""

    def retrieve_options(self, uargv, temp_dir):
        """Retrieve command line and option file options."""
        # convert command line arguments to string dictionary form
        remaining = self._get_arguments_dict(uargv)
        # unpack any packages
        package = self._parse_package_arg_and_unpack(remaining, temp_dir)
        # get preset groups from specified config file
        preset_dict = self._parse_config_arg_and_process_config_file(remaining)
        # parse default presets nested in config presets
        preset_dict = {
            _key: self._merge_arguments(
                self._parse_presets(_dict, PRESETS),
                _dict
            )
            for _key, _dict in iteritems(preset_dict)
        }
        # set defaults based on presets
        args = self._parse_presets(remaining, preset_dict)
        # local config file settings override preset settings
        self._merge_arguments(args, preset_dict[u'pcbasic'])
        # find unrecognised arguments
        unrecognised = ((_k, _v) for _k, _v in iteritems(args) if _k not in ARGUMENTS)
        for key, value in unrecognised:
            logging.warning(
                'Ignored unrecognised option `%s=%s` in configuration file', key, value
            )
        args = {_k: _v for _k, _v in iteritems(args) if _k in ARGUMENTS}
        # parse rest of command line args
        cmd_line_args = self._parse_args(remaining)
        # command-line args override config file settings
        self._merge_arguments(args, cmd_line_args)
        # parse GW-BASIC style options
        self._parse_gw_options(args)
        # clean up arguments
        self._convert_types(args)
        if package:
            # do not resume from a package
            args['resume'] = False
        return args

    def _append_short_args(self, args, key, value):
        """Append short arguments and value to dict."""
        # apply shorthands
        for short_arg, replacement in iteritems(SHORTHAND):
            key = key.replace(short_arg, replacement)
        long_arg_value = None
        for i, short_arg in enumerate(key[1:]):
            try:
                long_arg, long_arg_value = SHORT_ARGS[short_arg]
            except KeyError:
                logging.warning(u'Ignored unrecognised option `-%s`', short_arg)
            else:
                if i == len(key)-2:
                    # assign provided value to last argument specified
                    if long_arg_value and value:
                        logging.debug(
                            u'Value `%s` provided to option `-%s` interpreted as positional',
                            value, short_arg
                        )
                    self._append_arg(args, long_arg, long_arg_value or value or u'')
                else:
                    self._append_arg(args, long_arg, long_arg_value or u'')
        # if value provided not used, push back as positional
        if long_arg_value and value:
            return value
        return None

    def _append_arg(self, args, key, value):
        """Update a single list-type argument by appending a value."""
        if not value:
            # if we call _append_arg it means the key may be empty but is at least specified
            value = u''
        if key in args and args[key]:
            if value:
                args[key] += u',' + value
        else:
            args[key] = value

    def _get_arguments_dict(self, argv):
        """Convert command-line arguments to dictionary."""
        args = {}
        arg_deque = deque(argv)
        # positional arguments
        pos = 0
        # use -- to end option parsing, everything is a positional argument afterwards
        options_ended = False
        while arg_deque:
            arg = arg_deque.popleft()
            if not arg.startswith(u'-') or options_ended:
                # not an option flag, interpret as positional
                # strip enclosing quotes, but only if paired
                for quote in u'"\'':
                    if arg.startswith(quote) and arg.endswith(quote):
                        arg = arg.strip(quote)
                args[pos] = arg
                pos += 1
            elif arg == u'--':
                options_ended = True
            else:
                key, value = split_pair(arg, split_by=u'=', quote=u'"\'')
                # we know arg starts with -, not =, so key is not empty
                if key.startswith(u'--'):
                    # long option
                    if key[2:]:
                        self._append_arg(args, key[2:], value)
                else:
                    # starts with one dash
                    if not value:
                        # -key value, without = to connect
                        # only accept this for short options, long options with -- must have a =
                        # otherwise options with optional =True will absorb the option following
                        # only use the next value if it does not itself look like an option flag
                        if arg_deque:
                            if not arg_deque[0].startswith(u'-'):
                                value = arg_deque.popleft()
                    unused_value = self._append_short_args(args, key, value)
                    # if the value picked up is not used by the short option, push back as positional.
                    if unused_value:
                        arg_deque.appendleft(unused_value)
        return args

    def _parse_presets(self, remaining, conf_dict):
        """Parse presets"""
        presets = DEFAULT_SECTION
        try:
            argdict = {u'preset': remaining.pop(u'preset')}
        except KeyError:
            argdict = {}
        # apply default presets, including nested presets
        while True:
            # get dictionary of default config
            for p in presets:
                try:
                    self._merge_arguments(argdict, conf_dict[p])
                except KeyError:
                    if p not in DEFAULT_SECTION:
                        logging.warning(u'Ignored undefined preset `%s`', p)
            # look for more presets in expended arglist
            try:
                presets = self._to_list(u'preset', argdict.pop(u'preset'))
            except KeyError:
                break
        return argdict

    def _parse_package_arg_and_unpack(self, remaining, temp_dir):
        """Unpack zipfile package, if specified, and make its temp dir current."""
        # first positional arg: program or package name
        package = None
        try:
            arg_package = remaining[0]
        except KeyError:
            pass
        else:
            if os.path.isdir(arg_package):
                os.chdir(arg_package)
                remaining.pop(0)
                package = arg_package
            elif zipfile.is_zipfile(arg_package):
                remaining.pop(0)
                # extract the package to a temp directory
                # and make that the current dir for our run
                zipfile.ZipFile(arg_package).extractall(path=temp_dir)
                os.chdir(temp_dir)
                # if the zip-file contains only a directory at the top level,
                # then move into that directory. E.g. all files in package.zip
                # could be under the directory package/
                contents = os.listdir(u'.')
                if len(contents) == 1:
                    os.chdir(contents[0])
                # recursively rename all files to all-caps to avoid case issues on Unix
                # collisions: the last file renamed overwrites earlier ones
                for root, dirs, files in os.walk(u'.', topdown=False):
                    for name in dirs + files:
                        try:
                            os.rename(os.path.join(root, name), os.path.join(root, name.upper()))
                        except OSError:
                            # if we can't rename, ignore
                            pass
                package = arg_package
        # make package setting available
        return package

    def _parse_config_arg_and_process_config_file(self, remaining):
        """Find the correct config file and read it."""
        # always read default config files; private config overrides system config
        # we update a whole preset at once, there's no joining of settings.
        conf_dict = PRESETS.copy()
        conf_dict.update(self._read_config_file(USER_CONFIG_PATH))
        # find any local overriding config file & read it
        config_file = None
        try:
            config_file = remaining.pop(u'config')
        except KeyError:
            if os.path.exists(CONFIG_NAME):
                config_file = CONFIG_NAME
        if config_file:
            conf_dict.update(self._read_config_file(config_file))
        return conf_dict

    def _read_config_file(self, config_file):
        """Read config file."""
        try:
            config = configparser.RawConfigParser(allow_no_value=True)
            # use utf_8_sig to ignore a BOM if it's at the start of the file
            # (e.g. created by Notepad)
            with io.open(config_file, 'r', encoding='utf_8_sig', errors='replace') as f:
                if PY2: # pragma: no cover
                    config.readfp(WhitespaceStripper(f))
                else:
                    config.read_file(WhitespaceStripper(f))
        except (configparser.Error, IOError):
            logging.warning(
                u'Error in configuration file `%s`. Configuration not loaded.', config_file
            )
            return {u'pcbasic': {}}
        presets = {header: dict(config.items(header)) for header in config.sections()}
        return presets

    def _parse_args(self, remaining):
        """Process command line options."""
        # set arguments
        known = list(ARGUMENTS.keys()) + list(range(NUM_POSITIONAL))
        args = {d: remaining[d] for d in remaining if d in known}
        not_recognised = {d: remaining[d] for d in remaining if d not in known}
        for d in not_recognised:
            if not_recognised[d]:
                if isinstance(d, int):
                    logging.warning(
                        u'Ignored surplus positional command-line argument #%s: `%s`', d, not_recognised[d]
                    )
                else:
                    logging.warning(
                        u'Ignored unrecognised command-line argument `%s=%s`', d, not_recognised[d]
                    )
            else:
                logging.warning(u'Ignored unrecognised command-line argument `%s`', d)
        return args

    def _parse_gw_options(self, args):
        """Parse GW-BASIC-style options."""
        options = args.pop('options', '')
        for option in options.split(u' '):
            if not option:
                continue
            try:
                arg, val = GW_OPTIONS[option[:2].lower()], option[2:]
            except KeyError:
                try:
                    arg, val = GW_OPTIONS[option[:1]], option[1:]
                except KeyError:
                    # positional argument
                    arg, val = 0, option
            if val[:1] == u':':
                val = val[1:]
            arg, suffix = split_pair(arg, split_by=':', quote='"\'')
            args[arg] = val + (u':' + suffix if suffix else u'')
        return args


    ##########################################################################

    def _merge_arguments(self, target_dict, new_dict):
        """Update target_dict with new_dict. Lists of indefinite length are appended."""
        for a in new_dict:
            try:
                if (a in target_dict and ARGUMENTS[a][u'list'] == u'*' and target_dict[a]):
                    target_dict[a] += u',' + new_dict[a]
                    continue
            except KeyError:
                pass
            # override
            target_dict[a] = new_dict[a]
        return target_dict

    def _convert_types(self, args):
        """Convert arguments to required type and list length."""
        for name in args:
            try:
                args[name] = self._to_list(name, args[name], ARGUMENTS[name][u'list'])
            except KeyError:
                # not a list
                args[name] = self._parse_type(name, args[name])

    ##########################################################################
    # type conversions

    def _parse_type(self, d, arg):
        """Convert argument to required type."""
        if d not in ARGUMENTS:
            return arg
        if u'choices' in ARGUMENTS[d]:
            arg = arg.lower()
        first_arg = arg.split(u':')[0]
        if u'type' in ARGUMENTS[d]:
            if (ARGUMENTS[d][u'type'] == u'int'):
                arg = self._to_int(d, arg)
            elif (ARGUMENTS[d][u'type'] == u'bool'):
                arg = self._to_bool(d, arg)
        if u'choices' in ARGUMENTS[d]:
            if first_arg and first_arg not in ARGUMENTS[d][u'choices']:
                logging.warning(
                    u'Value `%s=%s` ignored; should be one of (`%s`)',
                    d, arg, u'`, `'.join(text_type(x) for x in ARGUMENTS[d][u'choices'])
                )
                arg = u''
        if u'check' in ARGUMENTS[d]:
            if arg and not ARGUMENTS[d][u'check'](first_arg):
                logging.warning(u'Value `%s=%s` ignored; not recognised', d, arg)
                arg = u''
        return arg

    def _to_list(self, argname, strval, length='*'):
        """Convert list strings to typed lists."""
        lst = strval.split(u',')
        if lst == [u'']:
            if length == '*':
                return []
            elif length < 0:
                return [None] * (-length)  # pylint: disable=invalid-unary-operand-type
            else:
                return None
        parsed = (self._parse_type(argname, _arg) for _arg in lst)
        lst = [_arg for _arg in parsed if _arg]
        # negative length: optional up-to
        if length != u'*' and length < 0:
            lst += [None] * (-length-len(lst))  # pylint: disable=invalid-unary-operand-type
        if length != u'*' and (len(lst) > abs(length) or len(lst) < length):
            logging.warning(
                u'Option `%s=%s` ignored: list should have %d elements, found %d',
                argname, strval, abs(length), len(lst)
            )
            lst = []
        if lst and argname in ARGUMENTS and u'listcheck' in ARGUMENTS[argname]:
            if not ARGUMENTS[argname][u'listcheck'](lst):
                logging.warning(u'Value "%s=%s" ignored; invalid', argname, strval)
                lst = []
        return lst

    def _to_bool(self, argname, strval):
        """Convert bool string to bool. Empty string (i.e. specified) means True."""
        if strval == u'':
            return True
        if strval.upper() in TRUES:
            return True
        elif strval.upper() in FALSES:
            return False
        else:
            logging.warning(
                u'Boolean option `%s=%s` interpreted as `%s=True`',
                argname, strval, argname
            )
        return True

    def _to_int(self, argname, strval):
        """Convert int string to int."""
        if strval:
            try:
                return int(strval)
            except ValueError:
                logging.warning(
                    u'Option `%s=%s` ignored: value should be an integer',
                    argname, strval
                )
        return None


##############################################################################
# utilities

class WhitespaceStripper(object):
    """File wrapper for ConfigParser that strips leading whitespace."""

    def __init__(self, file):
        """Initialise to file object."""
        self._file = file

    def readline(self):
        """Read a line and strip whitespace (but not EOL)."""
        return self._file.readline().lstrip(u' \t')

    def __next__(self):
        """Make iterable for Python 3."""
        line = self.readline()
        if not line:
            raise StopIteration()
        return line

    # iterability not actually needed in Python 2, but this keeps pylint happy
    next = __next__

    def __iter__(self):
        """We are iterable."""
        return self



"""
PC-BASIC - operators.py
Numeric and string operators

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ..base import tokens as tk
from .. import values


# operators and precedence
# key is tuple (token, nargs)
PRECEDENCE = {
    (tk.O_CARET, 2): 13,
    (tk.O_PLUS, 1): 12,
    (tk.O_MINUS, 1): 12,
    (tk.O_TIMES, 2): 11,
    (tk.O_DIV, 2): 11,
    (tk.O_INTDIV, 2): 10,
    (tk.MOD, 2): 9,
    (tk.O_PLUS, 2): 8,
    (tk.O_MINUS, 2): 8,
    (tk.O_GT, 2): 7,
    (tk.O_EQ, 2): 7,
    (tk.O_LT, 2): 7,
    (tk.O_GT + tk.O_EQ, 2): 7,
    (tk.O_EQ + tk.O_GT, 2): 7,
    (tk.O_LT + tk.O_EQ, 2): 7,
    (tk.O_EQ + tk.O_LT, 2): 7,
    (tk.O_LT + tk.O_GT, 2): 7,
    (tk.O_GT + tk.O_LT, 2): 7,
    (tk.NOT, 1): 6,
    (tk.AND, 2): 5,
    (tk.OR, 2): 4,
    (tk.XOR, 2): 3,
    (tk.EQV, 2): 2,
    (tk.IMP, 2): 1,
}
OPERATORS = set(operator_arity[0] for operator_arity in PRECEDENCE)

# can be combined like <> >=
COMBINABLE = (tk.O_LT, tk.O_EQ, tk.O_GT)

# unary operators
UNARY = {
    tk.O_MINUS: values.neg,
    tk.O_PLUS: lambda x: x,
    tk.NOT: values.not_,
}

# binary operators
BINARY = {
    tk.O_CARET: values.pow,
    tk.O_TIMES: values.mul,
    tk.O_DIV: values.div,
    tk.O_INTDIV: values.intdiv,
    tk.MOD: values.mod_,
    tk.O_PLUS: values.add,
    tk.O_MINUS: values.sub,
    tk.O_GT: values.gt,
    tk.O_EQ: values.eq,
    tk.O_LT: values.lt,
    tk.O_GT + tk.O_EQ: values.gte,
    tk.O_EQ + tk.O_GT: values.gte,
    tk.O_LT + tk.O_EQ: values.lte,
    tk.O_EQ + tk.O_LT: values.lte,
    tk.O_LT + tk.O_GT: values.neq,
    tk.O_GT + tk.O_LT: values.neq,
    tk.AND: values.and_,
    tk.OR: values.or_,
    tk.XOR: values.xor_,
    tk.EQV: values.eqv_,
    tk.IMP: values.imp_,
}



"""
PC-BASIC - interface.audio
Base class for audio plugins

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ..compat import queue
from ..basic.base import signals


class AudioPlugin(object):
    """Base class for audio interface plugins."""

    def __init__(self, audio_queue, **kwargs):
        """Setup the audio interface and start the event handling thread."""
        # sound generators for sounds not played yet
        # if not None, something is playing
        self._next_tone = [None, None, None, None]
        self.alive = True
        self._audio_queue = audio_queue

    # called by Interface

    @property
    def busy(self):
        """Something is playing."""
        return self._next_tone != [None, None, None, None]

    def cycle(self):
        """Audio event cycle."""
        if self.alive:
            self._drain_queue()
        if self.alive:
            self._work()

    # private methods

    def _drain_queue(self):
        """Drain audio queue."""
        while True:
            try:
                signal = self._audio_queue.get(False)
            except queue.Empty:
                return
            self._audio_queue.task_done()
            if signal.event_type == signals.QUIT:
                # close thread
                self.alive = False
            elif signal.event_type == signals.AUDIO_STOP:
                self.hush()
            elif signal.event_type == signals.AUDIO_PERSIST:
                self.persist(*signal.params)
            elif signal.event_type == signals.AUDIO_TONE:
                self.tone(*signal.params)
            elif signal.event_type == signals.AUDIO_NOISE:
                self.noise(*signal.params)

    # plugin overrides

    def __exit__(self, type, value, traceback):
        """Close the audio interface."""

    def __enter__(self):
        """Perform any necessary initialisations."""
        return self

    def _work(self):
        """Play some of the sounds queued."""

    # signal handlers

    def hush(self):
        """Be quiet."""

    def persist(self, do_persist):
        """Allow or disallow mixer to quit."""

    def tone(self, voice, frequency, duration, loop, volume):
        """Enqueue a tone."""

    def noise(self, source, frequency, duration, loop, volume):
        """Enqueue a noise."""



"""
PC-BASIC - cassette.py
Cassette Tape Device

(c) 2015--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import io
import math
import struct
import logging
from chunk import Chunk

from ...compat import int2byte, iterchar, zip

from ..base import error
from ..base import tokens as tk
from .devicebase import RawFile, TextFileBase, InputMixin, DeviceSettings, parse_protocol_string


# file types (data, bsaved memory, protected, ascii, tokenised)
TOKEN_TO_TYPE = {
    0: b'D', 1: b'M', 0xa0: b'P',
    0x20: b'P', 0x40: b'A', 0x80: b'B'
}

TYPE_TO_TOKEN = dict(reversed(item) for item in TOKEN_TO_TYPE.items())


#################################################################################
# Exceptions

class CassetteIOError(IOError): pass
class EndOfTape(CassetteIOError): pass
class CRCError(CassetteIOError): pass
class PulseError(CassetteIOError): pass
class FramingError(CassetteIOError): pass
class OperationNotImplemented(CassetteIOError): pass

#################################################################################
# Cassette device


class CASDevice(object):
    """Cassette tape device (CASn:) """

    # control characters not allowed in file name on tape
    _illegal_chars = set(int2byte(_i) for _i in range(0x20))

    def __init__(self, arg, console):
        """Initialise tape device."""
        addr, val = parse_protocol_string(arg)
        ext = val.split(u'.')[-1].upper()
        # WIDTH and LOC on CAS1: directly are ignored
        self.device_file = DeviceSettings()
        # by default, show messages
        self.is_quiet = False
        # console for messages
        self.console = console
        try:
            if not val:
                self.tapestream = None
            elif addr == u'WAV' or (addr != u'CAS' and ext == u'WAV'):
                # if unspecified, determine type on the basis of filename extension
                self.tapestream = CassetteStream(WAVBitStream(val, 'r'))
            else:
                # 'CAS' is default
                self.tapestream = CassetteStream(CASBitStream(val, 'r'))
        except EnvironmentError as e:
            logging.warning(u'Could not attach %s to CAS device: %s', val, e)
            self.tapestream = None

    def available(self):
        """Device is available."""
        return self.tapestream is not None

    def close(self):
        """Close tape device."""
        if self.tapestream:
            self.tapestream.close_tape()

    def open(self, number, param, filetype, mode, access, lock, reclen, seg, offset, length, field):
        """Open a file on tape."""
        if not self.tapestream:
            raise error.BASICError(error.DEVICE_UNAVAILABLE)
        if self.tapestream.is_open:
            raise error.BASICError(error.FILE_ALREADY_OPEN)
        if set(iterchar(param)) & self._illegal_chars:
            # Cassette BASIC throws bad file NUMBER, for some reason.
            raise error.BASICError(error.BAD_FILE_NUMBER)
        try:
            if mode == b'O':
                self.tapestream.open_write(param, filetype, seg, offset, length)
            elif mode == b'I':
                _, filetype, seg, offset, length = self._search(param, filetype)
            else:
                raise error.BASICError(error.BAD_FILE_MODE)
        except EnvironmentError:
            raise error.BASICError(error.DEVICE_IO_ERROR)
        if filetype == b'D':
            return CASTextFile(self.tapestream, filetype, mode)
        elif filetype == b'A':
            return CASTextFile(self.tapestream, filetype, mode)
        else:
            return CASBinaryFile(self.tapestream, filetype, mode, seg, offset, length)

    def _search(self, trunk_req=None, filetypes_req=None):
        """Play until a file header record is found for the given filename."""
        try:
            while True:
                trunk, filetype, seg, offset, length = self.tapestream.open_read()
                if (
                        (not trunk_req or trunk.rstrip() == trunk_req.rstrip()) and
                        (not filetypes_req or filetype in filetypes_req)
                    ):
                    message = b'%s.%s Found.' % (trunk, filetype)
                    if not self.is_quiet:
                        self.console.write_line(message)
                    logging.debug(timestamp(self.tapestream.counter()) + message)
                    return trunk, filetype, seg, offset, length
                else:
                    message = b'%s.%s Skipped.' % (trunk, filetype)
                    if not self.is_quiet:
                        self.console.write_line(message)
                    logging.debug(timestamp(self.tapestream.counter()) + message)
        except EndOfTape:
            # reached end-of-tape without finding appropriate file
            # we'll loop the tape for future use
            self.tapestream.wind(0)
            # timeout error to align with GW-BASIC behaviour
            raise error.BASICError(error.DEVICE_TIMEOUT)

    def quiet(self, is_quiet):
        """Suppress Skipped and Found messages."""
        self.is_quiet = is_quiet


#################################################################################
# Cassette files

class CASBinaryFile(RawFile):
    """Program or Memory file on CASn: device."""

    def __init__(self, fhandle, filetype, mode, seg, offset, length):
        """Initialise binary file."""
        RawFile.__init__(self, fhandle, filetype, mode)
        self.seg, self.offset, self.length = seg, offset, length


class CASTextFile(TextFileBase, InputMixin):
    """Text file on CASn: device."""

    def lof(self):
        """LOF: illegal function call."""
        raise error.BASICError(error.IFC)

    def loc(self):
        """LOC: illegal function call."""
        raise error.BASICError(error.IFC)

    def close(self):
        """Close a file on tape."""
        # terminate cassette text files with NUL
        if self.mode == b'O':
            self.write(b'\0')
        try:
            self._fhandle.close()
        except EnvironmentError:
            pass
        TextFileBase.close(self)


class CassetteStream(object):
    """Byte stream on CASn: device."""

    def __init__(self, bitstream):
        """Initialise file on tape."""
        self.bitstream = bitstream
        # is a file open on this stream?
        self.is_open = False
        # keep track of last seg, offs, length to reproduce GW-BASIC oddity
        self.last = 0, 0, 0
        self.buffer_complete = False
        self.length = 0
        self.filetype = b''
        self.rwmode = ''

    def close(self):
        """Finalise the track on the tape stream."""
        if self.is_open:
            self._close_record_buffer()
            self.is_open = False
            self.rwmode = ''

    def close_tape(self):
        """Eject the tape."""
        try:
            self.close()
            self.bitstream.close()
        except EnvironmentError:
            pass

    def counter(self):
        """Position on tape in seconds."""
        return self.bitstream.counter()

    def wind(self, loc):
        self.bitstream.wind(loc)

    def write(self, c):
        """Write a string to a file on tape."""
        self.record_stream.write(c)
        self._flush_record_buffer()

    def read(self, nbytes=-1):
        """Read bytes from a file on tape."""
        c = b''
        try:
            while True:
                if nbytes > -1:
                    c += self.record_stream.read(nbytes-len(c))
                    if len(c) >= nbytes:
                        return c
                else:
                    c += self.record_stream.read()
                if self.buffer_complete:
                    return c
                self._fill_record_buffer()
        except EndOfTape:
            return c

    def open_read(self):
        """Play until a file header record is found."""
        self.record_num = 0
        self.record_stream = io.BytesIO()
        self.buffer_complete = False
        self.bitstream.switch_mode('r')
        self.rwmode = 'r'
        while True:
            record = self._read_record(None)
            if record and record[0:1] == b'\xa5':
                break
            else:
                # unknown record type
                logging.debug('%s Skipped non-header record.', timestamp(self.bitstream.counter()))
        file_trunk, token, self.length, seg, offset = struct.unpack('<8sBHHH', record[1:16])
        try:
            self.filetype = TOKEN_TO_TYPE[token]
        except KeyError:
            logging.debug('Unknown file type token: %x', token)
        self.record_num = 0
        self.buffer_complete = False
        self.is_open = True
        return file_trunk, self.filetype, seg, offset, self.length

    def open_write(self, name, filetype, seg, offs, length):
        """Write a file header to the tape."""
        self.record_num = 0
        self.record_stream = io.BytesIO()
        self.buffer_complete = False
        self.bitstream.switch_mode('w')
        self.rwmode = 'w'
        if filetype in (b'A', b'D'):
            # ASCII program files: length, seg, offset are untouched,
            # remain that of the previous file recorded!
            seg, offs, length = self.last
        else:
            self.last = seg, offs, length
        self.filetype = filetype
        # header seems to end at 0x00, 0x01, then filled out with last char
        header = struct.pack(
            '<c8sBHHHBB',
            b'\xa5', name[:8] + b' ' * (8-len(name)),
            TYPE_TO_TOKEN[filetype], length, seg, offs, 0, 1
        )
        self._write_record(header)
        self.is_open = True

    def _read_record(self, reclen):
        """Read a record from tape."""
        if not self.bitstream.read_leader():
            # reached end-of-tape without finding appropriate file
            raise EndOfTape()
        self.record_num += 1
        record = b''
        block_num = 0
        byte_count = 0
        while reclen is None or byte_count < reclen:
            data = self._read_block()
            record += data
            byte_count += len(data)
            if (reclen is None):
                break
            block_num += 1
        self.bitstream.read_trailer()
        if reclen is not None:
            return record[:reclen]
        return record

    def _write_record(self, data):
        """Write a data record to tape."""
        self.bitstream.write_leader()
        while len(data) > 0:
            self._write_block(data[:256])
            data = data[256:]
        self.bitstream.write_trailer()
        # write 100 ms pause to make clear separation between blocks
        self.bitstream.write_pause(100)

    def _read_block(self):
        """Read a block of data from tape."""
        count = 0
        data = b''
        while True:
            if count == 256:
                break
            byte = self.bitstream.read_byte()
            if byte is None:
                raise PulseError()
            data += int2byte(byte)
            count += 1
        bytes0, bytes1 = self.bitstream.read_byte(), self.bitstream.read_byte()
        crc_given = bytes0 * 0x100 + bytes1
        crc_calc = crc(data)
        # if crc for either polarity matches, return that
        if crc_given == crc_calc:
            return data
        raise CRCError('CRC check failed, required: %04x realised: %04x' % (crc_given, crc_calc))

    def _write_block(self, data):
        """Write a 256-byte block to tape."""
        # fill out short blocks with last byte
        data += data[-1:]*(256-len(data))
        for b in iterchar(data):
            self.bitstream.write_byte(ord(b))
        crc_word = crc(data)
        # crc is written big-endian
        lo, hi = (ord(_b) for _b in iterchar(struct.pack('<H', crc_word)))
        self.bitstream.write_byte(hi)
        self.bitstream.write_byte(lo)

    def _fill_record_buffer(self):
        """Read to fill the tape buffer."""
        if self.buffer_complete:
            return False
        if self.filetype in (b'M', b'B', b'P'):
            # bsave, tokenised and protected come in one multi-block record
            self.record_stream = io.BytesIO()
            self.record_stream.write(self._read_record(self.length))
            self.buffer_complete = True
        else:
            # ascii and data come as a sequence of one-block records
            # 256 bytes less 1 length byte. CRC trailer comes after 256-byte block
            self.record_stream = io.BytesIO()
            record = self._read_record(256)
            num_bytes = ord(record[0:1])
            record = record[1:]
            if num_bytes != 0:
                record = record[:num_bytes-1]
                self.buffer_complete = True
            self.record_stream.write(record)
        self.record_stream.seek(0)
        return True

    def _flush_record_buffer(self):
        """Write the tape buffer to tape."""
        if self.filetype not in (b'M', b'B', b'P') and self.rwmode == 'w':
            data = self.record_stream.getvalue()
            while True:
                if len(data) < 255:
                    break
                chunk, data = data[:255], data[255:]
                # ascii and data come as a sequence of one-block records
                # 256 bytes less 1 length byte. CRC trailer comes after 256-byte block
                self._write_record(b'\0' + chunk)
            self.record_stream = io.BytesIO()
            self.record_stream.write(data)
            self.record_stream.seek(0, 2)

    def _close_record_buffer(self):
        """Write the tape buffer to tape and finalise."""
        if self.rwmode == 'w':
            self._flush_record_buffer()
            self.buffer_complete = True
            data = self.record_stream.getvalue()
            if self.filetype in (b'M', b'B', b'P'):
                # bsave, tokenised and protected come in one multi-block record
                self._write_record(data)
            else:
                if data:
                    self._write_record(int2byte(len(data)) + data)
        self.record_stream = io.BytesIO()


##############################################################################


class TapeBitStream(object):
    """Cassette tape bitstream interface."""

    # sync byte for IBM PC tapes
    sync_byte = 0x16
    # intro text
    intro = b'PC-BASIC tape\x1a'

    def __init__(self, mode='r'):
        """Initialise tape interface."""
        pass

    def __enter__(self):
        """Context guard."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Context guard."""
        self.close()

    def counter(self):
        """Position on tape in seconds."""
        return 0

    def wind(self, loc):
        """Set position of tape in seconds."""
        pass

    def read_intro(self):
        """Try to read intro; ensure image not empty."""
        for b in bytearray(self.intro):
            try:
                c = self.read_byte()
            except EndOfTape:
                # empty or short file
                return False
            if c != b:
                break
        else:
            for _ in range(7):
                self.read_bit()
        return True

    def write_intro(self):
        """Write some noise to give the reader something to get started."""
        # We just need some bits here
        # however on a new CAS file this works like a magic-sequence...
        for b in bytearray(self.intro):
            self.write_byte(b)
        # Write seven bits, so that we are byte-aligned after the sync bit
        # (after the 256-byte pilot). Makes CAS-files easier to read in hex.
        for _ in range(7):
            self.write_bit(0)
        self.write_pause(100)

    def read_leader(self):
        """Read the leader / pilot wave."""
        try:
            while True:
                while self.read_bit() != 1:
                    pass
                counter = 0
                while True:
                    b = self.read_bit()
                    if b != 1:
                        break
                    counter += 1
                # sync bit 0 has been read, check sync byte 0x16
                # at least 64*8 bits
                if b is not None and counter >= 512:
                    sync = self.read_byte(skip_start=True)
                    if sync == self.sync_byte:
                        return True
        except EndOfTape:
            return False

    def write_leader(self):
        """Write the leader / pilot tone."""
        for _ in range(256):
            self.write_byte(0xff)
        self.write_bit(0)
        self.write_byte(0x16)

    def read_byte(self, skip_start=False):
        """Read a byte from the tape."""
        # NOTE: skip_start is ignored
        byte = 0
        for i in range(8):
            bit = self.read_bit()
            if bit is None:
                raise PulseError()
            byte += bit * 128 >> i
        return byte

    def write_byte(self, byte):
        """Write a byte to tape image."""
        bits = [1 if (byte & (128 >> i) != 0) else 0 for i in range(8)]
        for bit in bits:
            self.write_bit(bit)

    def close(self):
        """Eject tape."""
        pass

    def switch_mode(self, new_mode):
        """Switch tape to reading or writing mode."""
        pass

    def flush(self):
        """Write remaining bits to tape (stub)."""
        pass

    def read_bit(self):
        """Read the next bit (stub)."""
        return 0

    def write_bit(self, bit):
        """Write the next bit (stub)."""
        pass

    def write_pause(self, milliseconds):
        """Write pause to tape image (stub)."""
        pass

    def read_trailer(self):
        """Read the trailing wave """
        try:
            while self.read_bit() == 1:
                pass
        except EndOfTape:
            pass

    def write_trailer(self):
        """Write trailing wave."""
        # closing sequence is 30 1-bits followed by a zero bit (based on PCE output).
        # Not 32 1-bits as per http://fileformats.archiveteam.org/wiki/IBM_PC_data_
        for _ in range(30):
            self.write_bit(1)
        self.write_bit(0)

##############################################################################


class CASBitStream(TapeBitStream):
    """CAS-file cassette image bit stream."""

    def __init__(self, image_name, mode):
        """Initialise CAS-file."""
        TapeBitStream.__init__(self)
        # 'r' or 'w'
        self.cas_name = image_name
        if not os.path.exists(self.cas_name):
            self._create()
        else:
            self.operating_mode = 'r'
            self.mask = 0x100
            try:
                self.cas = io.open(self.cas_name, 'r+b')
            except EnvironmentError:
                self.cas = io.open(self.cas_name, 'rb')
            self.current_byte = self.cas.read(1)
            if self.current_byte == b'' or not self.read_intro():
                self.cas.close()
                self._create()
        self.switch_mode(mode)

    def __getstate__(self):
        """Get pickling dict for stream."""
        return {
            'filename': self.cas_name,
            'mode': self.operating_mode,
            'counter': self.counter()
        }

    def __setstate__(self, st):
        """Initialise stream from pickling dict."""
        self.__init__(st['filename'], st['mode'])
        self.wind(st['counter'])

    def counter(self):
        """Time stamp in seconds."""
        # approximate: average 750 us per bit, cut on bytes
        return self.cas.tell() * 8 * 750 / 1000000.

    def wind(self, loc):
        """Set position of tape in seconds."""
        self.cas.seek(int(loc * 1000000 // (750 * 8)))
        self.current_byte = self.cas.read(1)

    def close(self):
        """Close tape image."""
        # ensure any buffered bits are written
        self.flush()
        self.cas.close()

    def read_bit(self):
        """Read the next bit."""
        self.mask >>= 1
        if self.mask <= 0:
            self.current_byte = self.cas.read(1)
            if not self.current_byte:
                raise EndOfTape()
            self.mask = 0x80
        if (ord(self.current_byte) & self.mask == 0):
            return 0
        else:
            return 1

    def write_bit(self, bit):
        """Write a bit to tape."""
        # note that CAS-files aren't necessarily byte aligned
        # the ones we make are, but PCE's ones aren't.
        self.mask >>= 1
        if self.mask <= 0:
            self.cas.write(self.current_byte)
            self.current_byte = b'\0'
            self.mask = 0x80
        self.current_byte = int2byte(ord(self.current_byte) | (bit*self.mask))

    def flush(self):
        """Write remaining bits to tape."""
        if self.operating_mode == 'w':
            # write -> read
            # read bit on stream to combine with
            existing = self.cas.read(1)
            if not existing:
                existing = b'\0'
            else:
                self.cas.seek(-1, 1)
            # 0b1000 -> 0b1111 etc.
            combine_mask = self.mask * 2 - 1
            self.current_byte = int2byte(
                (ord(existing) & combine_mask) +
                (ord(self.current_byte) & (0xff^combine_mask))
            )
            # flush bits in write buffer
            # pad with zero if necessary to align on byte limit
            self.cas.write(self.current_byte)
            # if we continue to write, we should seek(-1,1)
            self.cas.seek(-1, 1)

    def switch_mode(self, new_mode):
        """Switch tape to reading or writing mode."""
        self.flush()
        if self.operating_mode == 'w' and new_mode == 'r':
            self.current_byte = self.cas.read(1)
        elif self.operating_mode == 'r' and new_mode == 'w':
            self.cas.seek(-1, 1)
        self.operating_mode = new_mode

    def _create(self):
        """Create a new CAS-file."""
        self.current_byte = b'\0'
        self.mask = 0x100
        with io.open(self.cas_name, 'wb') as self.cas:
            self.operating_mode = 'w'
            self.current_byte = b'\0'
            self.write_intro()
        self.cas = io.open(self.cas_name, 'r+b')
        self.cas.seek(0, 2)



# http://www.topherlee.com/software/pcm-tut-wavformat.html
# The header of a WAV (RIFF) file is 44 bytes long and has the following format:

#    Positions	Sample Value	Description
#    1 - 4	"RIFF"	Marks the file as a riff file. Characters are each 1 byte long.
#    5 - 8	File size (integer)	Size of the overall file - 8 bytes, in bytes (32-bit integer). Typically, you'd fill this in after creation.
#    9 -12	"WAVE"	File Type Header. For our purposes, it always equals "WAVE".
#    13-16	"fmt "	Format chunk marker. Includes trailing null
#    17-20	16	Length of format data as listed above
#    21-22	1	Type of format (1 is PCM) - 2 byte integer
#    23-24	2	Number of Channels - 2 byte integer
#    25-28	44100	Sample Rate - 32 byte integer. Common values are 44100 (CD), 48000 (DAT). Sample Rate = Number of Samples per second, or Hertz.
#    29-32	176400	(Sample Rate * BitsPerSample * Channels) / 8.
#    33-34	4	(BitsPerSample * Channels) / 8.1 - 8 bit mono2 - 8 bit stereo/16 bit mono4 - 16 bit stereo
#    35-36	16	Bits per sample
#    37-40	"data"	"data" chunk header. Marks the beginning of the data section.
#    41-44	File size (data)	Size of the data section.
#    Sample values are given above for a 16-bit stereo source.

# data section consists of little-endian PCM audio data
# each sample consists of nchannels*samplewidth bytes


# pc cassette technical docs
# http://en.wikipedia.org/wiki/IBM_cassette_tape
# http://www.vintage-computer.com/vcforum/showthread.php?8829-IBM-PC-Cassette-interface/page2


# http://fileformats.archiveteam.org/wiki/IBM_PC_data_cassette
# The format consists of 1-millisecond-long pulses for each 1 bit, and
# 0.5-millisecond pulses for each 0 bit. A tape record starts with a leader
# of 256 bytes of all 1 bits (hex FF), followed by a single synchronization
# bit (0), and then a synchronization byte (hex 16, the ASCII character from
# the C0 controls designated as SYN). This is followed by one or more 256-byte
# data blocks. Each data block is followed by a 2-byte CRC, with the most
# significant byte first. After the last block, a 4-byte trailer is written
# of all 1 bits (hex FF).
#
# NOTE that the trailer consists of 30 1 bits and a 0 bit, not 32 1-bits as described here.

# Tokenised BASIC programs and memory areas saved by IBM Cassette BASIC consist
# of two records: the first one is a header (always 256 bytes, of which the
# first 16 are significant), and the second one contains the data.
# ASCII listings and data files consist of a sequence of 256-byte records; the
# first one is a header, as above, and subsequent ones contain the data. If
# the first byte of the record is 0, this is not the last record, and all 255
# following bytes are valid data. Otherwise it gives the number of valid bytes
# in the last record, plus one.
# The header layout is:
# Offset	Size	Description
# 0x00	Byte	Always 0xA5
# 0x01	8 bytes	Filename, ASCII
# 0x09	Byte	Flags:
# Bit	Meaning if set	Example command to create
# 7	Tokenised BASIC	SAVE "file"
# 6	ASCII listing	SAVE "file", A
# 5	Protected tokenised BASIC	SAVE "file", P
# 0	Memory area	BSAVE "file", address, length
# No bits set	Data	OPEN "O",1,"file"
# 0x0A	Word	Number of bytes in the following data record (little-endian word)
# 0x0C	Word	Segment of load address (little-endian word)
# 0x0E	Word	Offset of load address (little-endian word)


class WAVBitStream(TapeBitStream):
    """WAV-file cassette image bit stream."""

    def __init__(self, filename, mode):
        """Initialise WAV-file."""
        TapeBitStream.__init__(self)
        self.filename = filename
        if not os.path.exists(filename):
            # create/overwrite file
            self.framerate = 22050
            self.sampwidth = 1
            self.nchannels = 1
            self.wav = io.open(self.filename, 'wb')
            self._write_wav_header()
            self.operating_mode = 'w'
        else:
            # open file for reading and find wave parameters
            try:
                self.wav = io.open(self.filename, 'r+b')
            except EnvironmentError:
                self.wav = io.open(self.filename, 'rb')
            if not self._read_wav_header():
                raise EndOfTape()
            self.operating_mode = 'r'
        self.wav_pos = 0
        self.buf_len = 1024
        # convert 8-bit and 16-bit values to ints
        if self.sampwidth == 1:
            self.sub_threshold = 0
            self.subtractor = 128*self.nchannels
        else:
            self.sub_threshold = 256*self.nchannels//2
            self.subtractor =  256*self.nchannels
        # volume above/below zero that is interpreted as zero
        self.zero_threshold = self.nchannels
        # 1000 us for 1, 500 us for 0; threshold for half-pulse (500 us, 250 us)
        self.halflength = [(250*self.framerate) // 1000000, (500*self.framerate) // 1000000]
        self.halflength_cut = (375 * self.framerate) // 1000000
        self.halflength_max = 2 * self.halflength_cut
        self.halflength_min = self.halflength_cut // 2
        self.length_cut = 2*self.halflength_cut
        # 2048 halves = 1024 pulses = 512 1-bits = 64 bytes of leader
        self.min_leader_halves = 2048
        # initialise generators
        self.filter = passthrough()
        self.filter.send(None)
        self.read_half = self._gen_read_halfpulse()
        # write fluff at start if this is a new file
        if self.operating_mode == 'w':
            self.write_intro()
        self.switch_mode(mode)

    def __getstate__(self):
        """Get pickling dict for stream."""
        return {
            'filename': self.filename,
            'mode': self.operating_mode,
            'counter': self.counter()
        }

    def __setstate__(self, st):
        """Initialise stream from pickling dict."""
        # open for reading to avoid writing intro
        self.__init__(st['filename'], 'r')
        self.wind(st['counter'])
        self.switch_mode(st['mode'])

    def switch_mode(self, mode):
        """Switch tape to reading or writing mode."""
        self.operating_mode = mode

    def counter(self):
        """Time stamp in seconds."""
        return self.wav_pos/(1.*self.framerate)

    def wind(self, loc):
        """Set position of tape in seconds."""
        self.wav_pos = int(loc * self.framerate)
        self.wav.seek(self.wav_pos)

    def read_bit(self):
        """Read the next bit."""
        try:
            length_up, length_dn = next(self.read_half), next(self.read_half)
        except StopIteration:
            self.read_half = self._gen_read_halfpulse()
            raise EndOfTape()
        if (length_up > self.halflength_max or length_dn > self.halflength_max or
                length_up < self.halflength_min or length_dn < self.halflength_min):
            return None
        elif length_up >= self.halflength_cut:
            return 1
        else:
            return 0

    def close(self):
        """Close WAV-file."""
        TapeBitStream.close(self)
        # write file length fields
        self.wav.seek(0, 2)
        end_pos = self.wav.tell()
        self.wav.seek(self.riff_pos, 0)
        self.wav.write(struct.pack('<4sL', b'RIFF', end_pos-self.riff_pos-8))
        self.wav.seek(self.data_pos, 0)
        self.wav.write(struct.pack('<4sL', b'data', end_pos-self.start))
        self.wav.close()

    def _fill_buffer(self):
        """Fill buffer with frames and pre-process."""
        frame_buf = []
        frames = self.wav.read(self.buf_len*self.nchannels*self.sampwidth)
        if not frames:
            raise EndOfTape
        # convert MSBs to int (data stored little endian)
        # note that we simply throw away all the less significant bytes
        frames = (ord(_c) for _c in iterchar(frames[self.sampwidth-1::self.sampwidth]))
        # sum frames over channels
        frames = map(sum, zip(*[iter(frames)]*self.nchannels))
        frames = [
            _x-self.subtractor if _x >= self.sub_threshold else _x
            for _x in frames
        ]
        return self.filter.send(frames)

    def _gen_read_halfpulse(self):
        """Generator to read a half-pulse and yield its length."""
        length = 0
        frame = 1
        prezero = 1
        pos_in_frame = 0
        frame_buf = []
        while True:
            try:
                sample = frame_buf[pos_in_frame]
                pos_in_frame += 1
            except IndexError:
                frame_buf = self._fill_buffer()
                pos_in_frame = 0
                continue
            length += 1
            last, frame = frame, (sample > self.zero_threshold) + (sample >= -self.zero_threshold) - 1
            if last != frame and (last != 0 or frame == prezero):
                if frame == 0 and last != 0:
                    prezero = last
                self.wav_pos += length
                yield length
                length = 0

    def write_pause(self, milliseconds):
        """Write a pause of given length to the tape."""
        length = int(milliseconds * self.framerate / 1000)
        zero = {1: b'\x7f', 2: b'\x00\x00'}
        self.wav.write(zero[self.sampwidth] * self.nchannels * length)
        self.wav_pos += length

    def write_bit(self, bit):
        """Write a bit to tape."""
        half_length = self.halflength[bit]
        down = {1: b'\x00', 2: b'\x00\x80'}
        up = {1: b'\xff', 2: b'\xff\x7f'}
        self.wav.write(
            down[self.sampwidth] * self.nchannels * half_length +
            up[self.sampwidth] * self.nchannels * half_length
        )
        self.wav_pos += 2 * half_length

    def _read_wav_header(self):
        """Read RIFF WAV header."""
        try:
            ch = Chunk(self.wav, bigendian=0)
        except (EOFError):
            logging.debug('WAV file is corrupted.')
            return False
        if ch.getname() != b'RIFF' or ch.read(4) != b'WAVE':
            logging.debug('Not a WAV file.')
            return False
        # this would normally be 0
        self.riff_pos = self.wav.tell() - 12
        riff_size = ch.getsize()
        self.sampwidth, self.nchannels, self.framerate = 0, 0, 0
        while True:
            try:
                chunk = Chunk(ch, bigendian=0)
            except EOFError:
                logging.debug('No data chunk found in WAV file.')
                return False
            chunkname = chunk.getname()
            if chunkname == b'fmt ':
                format_tag, self.nchannels, self.framerate, _, _ = struct.unpack(
                    '<HHLLH', chunk.read(14)
                )
                if format_tag == 1:
                    sampwidth = struct.unpack('<H', chunk.read(2))[0]
                    self.sampwidth = (sampwidth + 7) // 8
                else:
                    logging.debug('WAV file not in uncompressed PCM format.')
                    return False
            elif chunkname == b'data':
                if not self.sampwidth:
                    logging.debug('Format chunk not found.')
                    return False
                self.data_pos = self.wav.tell() - 4
                #self.wav.read(4)
                self.start = self.wav.tell()
                return True
            chunk.skip()

    def _write_wav_header(self):
        """Write RIFF WAV header."""
        # "RIFF" chunk header
        self.riff_pos = self.wav.tell()
        # length is corrected at close
        self.wav.write(struct.pack('<4sL4s', b'RIFF', 36, b'WAVE'))
        # "fmt " subchunk
        self.wav.write(struct.pack(
            '<4sLHHLLHH',
            b'fmt ', 16, 1, self.nchannels, self.framerate,
            self.nchannels * self.framerate * self.sampwidth,
            self.nchannels * self.sampwidth,
            self.sampwidth * 8
        ))
        # "data" subchunk header
        self.data_pos = self.wav.tell()
        # length is corrected at close
        self.wav.write(struct.pack('<4sL', b'data', 0))
        self.start = self.wav.tell()

    def _is_leader_halfpulse(self, half):
        """Return whether the half pulse is of pilot wave frequency."""
        return half >= self.length_cut/2

    def read_leader(self):
        """Read the leader / pilot wave."""
        try:
            while True:
                while self.read_bit() != 1:
                    pass
                counter = 0
                pulse = (0,0)
                while True:
                    last = pulse
                    half = next(self.read_half)
                    if not self._is_leader_halfpulse(half):
                        if counter > self.min_leader_halves:
                            #  zero bit; try to sync
                            half = next(self.read_half)
                        break
                    counter += 1
                # sync bit 0 has been read, check sync byte
                if counter >= self.min_leader_halves:
                    # read rest of first byte
                    try:
                        self.last_error_bit = None
                        self.dropbit = None
                        sync = self.read_byte(skip_start=True)
                        if sync == self.sync_byte:
                            return True
                        else:
                            logging.debug(
                                '%s Incorrect sync byte after %d pulses: %02x',
                                timestamp(self.counter()), counter, sync
                            )
                    except (PulseError, FramingError) as e:
                        logging.debug(
                            '%s Error in sync byte after %d pulses: %s',
                            timestamp(self.counter()), counter, e
                        )
        except (EndOfTape, StopIteration):
            self.read_half = self._gen_read_halfpulse()
            return False

##############################################################################
# supporting functions

def crc(data):
    """Calculate 16-bit CRC-16-CCITT for data."""
    # see http://en.wikipedia.org/wiki/Computation_of_cyclic_redundancy_checks
    # for a lookup table version, see e.g. WAV2CAS v1.3 for Poisk PC. by Tronix (C) 2013
    # however, speed is not critical for this function
    rem = 0xffff
    for d in bytearray(data):
        rem ^= d << 8
        for _ in range(8):
            rem <<= 1
            if rem & 0x10000:
                rem ^= 0x1021
            rem &= 0xffff
    return rem ^ 0xffff

def hms(seconds):
    """Return elapsed cassette time at given frame."""
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return h, m, s

def timestamp(counter):
    """Time stamp."""
    return b'[%d:%02d:%02d] ' % hms(counter)

def passthrough():
    """Passthrough filter."""
    x = []
    while True:
        x = yield x



"""
PC-BASIC - protect.py
Source encryption/decryption

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.

Paul Kocher reverse engineered the decryption algorithm, published in:
The Cryptogram computer supplement #19, American Cryptogram Association, Summer 1994
"""

from ...compat import int2byte


# 13-byte and 11-byte keys used by GW-BASIC
KEY1 = (0xA9, 0x84, 0x8D, 0xCD, 0x75, 0x83, 0x43, 0x63, 0x24, 0x83, 0x19, 0xF7, 0x9A)
KEY2 = (0x1E, 0x1D, 0xC4, 0x77, 0x26, 0x97, 0xE0, 0x74, 0x59, 0x88, 0x7C)


def unprotect(ins, outs):
    """Decrypt a byte stream read from the GWBASIC ,P (read protected) format. This will allow it to be subsequently parsed."""
    index = 0
    s = ins.read(1)
    while s != b'':
        nxt = ins.read(1)
        # drop last char (EOF 0x1a)
        if nxt == b'':
            break
        c = ord(s)
        # Kocher's algorithm:
        c -= 11 - (index % 11)
        c ^= KEY1 [index % 13]
        c ^= KEY2 [index % 11]
        c += 13 - (index % 13)
        outs.write(int2byte(c % 256))
        index = (index+1) % (13*11)
        s = nxt
    # return last char written
    return int2byte(c % 256)

def protect(ins, outs):
    """Encrypt a byte stream read from the GWBASIC tokenised format."""
    index = 0
    nxt = ins.read(1)
    while nxt != b'':
        s = nxt
        c = ord(s)
        # inverse Kocher's algorithm:
        c -= 13 - (index % 13)
        c ^= KEY1 [index % 13]
        c ^= KEY2 [index % 11]
        c += 11 - (index % 11)
        outs.write(int2byte(c % 256))
        index = (index+1) % (13*11)
        nxt = ins.read(1)
    # return last char read
    return s



"""
PC-BASIC - interface.audio_portaudio
Sound interface based on PortAudio

(c) 2015--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os
import logging
from collections import deque
from contextlib import contextmanager

if False:
    # for detection by packagers
    import pyaudio

from ..compat import zip, WIN32, MACOS
from .audio import AudioPlugin
from .base import audio_plugins, InitFailed
from . import synthesiser


# approximate generator chunk length
# one wavelength at 37 Hz is 1192 samples at 44100 Hz
_CHUNK_LENGTH = 1192 * 4
# buffer size in sample frames
_BUFSIZE = 1024


# suppress ALSA debug messages
# https://stackoverflow.com/questions/7088672/pyaudio-working-but-spits-out-error-messages-each-time
if WIN32 or MACOS:
    @contextmanager
    def _quiet_alsa(): yield
else:
    @contextmanager
    def _quiet_alsa():
        """Set the error handler in alsa to avoid debug messages on the screen."""
        from ctypes import CFUNCTYPE, c_char_p, c_int, cdll
        # From alsa-lib Git 3fd4ab9be0db7c7430ebd258f2717a976381715d
        # $ grep -rn snd_lib_error_handler_t
        # include/error.h:59:typedef void (*snd_lib_error_handler_t)(const char *file, int line, const char *function, int err, const char *fmt, ...) /* __attribute__ ((format (printf, 5, 6))) */;
        ERROR_HANDLER_FUNC = CFUNCTYPE(None, c_char_p, c_int, c_char_p, c_int, c_char_p)
        def py_error_handler(filename, line, function, err, fmt):
            logging.debug('ALSA: %s:%s: %s', filename, line, fmt)
        c_error_handler = ERROR_HANDLER_FUNC(py_error_handler)
        asound = cdll.LoadLibrary('libasound.so')
        asound.snd_lib_error_set_handler(c_error_handler)
        try:
            yield
        finally:
            asound.snd_lib_error_set_handler(None)



@audio_plugins.register('portaudio')
class AudioPortAudio(AudioPlugin):
    """PortAudio-based audio plugin."""

    def __init__(self, audio_queue, **kwargs):
        """Initialise sound system."""
        global pyaudio
        try:
            import pyaudio
        except ImportError:
            raise InitFailed('Module `pyaudio` not found')
        # synthesisers
        self._signal_sources = synthesiser.get_signal_sources()
        # sound generators for each voice
        self._generators = [deque() for _ in synthesiser.VOICES]
        # buffer of samples; drained by callback, replenished by _play_sound
        self._samples = [bytearray() for _ in synthesiser.VOICES]
        self._device = None
        self._stream = None
        self._alsa_muffle = None
        self._min_samples_buffer = 2 * _BUFSIZE
        AudioPlugin.__init__(self, audio_queue)

    def __enter__(self):
        """Perform any necessary initialisations."""
        self._alsa_muffle = _quiet_alsa()
        self._alsa_muffle.__enter__()
        self._device = pyaudio.PyAudio()
        self._stream = self._device.open(
            format=pyaudio.paInt8, channels=1, rate=synthesiser.SAMPLE_RATE, output=True,
            frames_per_buffer=_BUFSIZE, stream_callback=self._get_next_chunk
        )
        self._stream.start_stream()
        AudioPlugin.__enter__(self)

    def __exit__(self, type, value, traceback):
        """Close down PortAudio."""
        self._alsa_muffle.__exit__(type, value, traceback)
        self._stream.stop_stream()
        self._stream.close()
        self._device.terminate()
        return AudioPlugin.__exit__(self, type, value, traceback)

    def tone(self, voice, frequency, duration, loop, volume):
        """Enqueue a tone."""
        self._generators[voice].append(synthesiser.SoundGenerator(
            self._signal_sources[voice], synthesiser.FEEDBACK_TONE,
            frequency, duration, loop, volume
        ))

    def noise(self, source, frequency, duration, loop, volume):
        """Enqueue a noise."""
        feedback = synthesiser.FEEDBACK_NOISE if source else synthesiser.FEEDBACK_PERIODIC
        self._generators[synthesiser.NOISE_VOICE].append(synthesiser.SoundGenerator(
            self._signal_sources[synthesiser.NOISE_VOICE], feedback,
            frequency, duration, loop, volume
        ))

    def hush(self):
        """Stop sound."""
        self._next_tone = [None for _ in synthesiser.VOICES]
        for gen in self._generators:
            gen.clear()
        self._samples = [bytearray() for _ in synthesiser.VOICES]

    def _work(self):
        """Replenish sample buffer."""
        for voice in synthesiser.VOICES:
            if len(self._samples[voice]) > self._min_samples_buffer:
                # nothing to do
                continue
            while True:
                if self._next_tone[voice] is None or self._next_tone[voice].loop:
                    try:
                        # looping tone will be interrupted
                        # by any new tone appearing in the generator queue
                        self._next_tone[voice] = self._generators[voice].popleft()
                    except IndexError:
                        if self._next_tone[voice] is None:
                            current_chunk = None
                            break
                current_chunk = self._next_tone[voice].build_chunk(_CHUNK_LENGTH)
                if current_chunk is not None:
                    break
                self._next_tone[voice] = None
            if current_chunk is not None:
                # append chunk to samples list
                # should lock to ensure callback doesn't try to access the list too?
                self._samples[voice] = bytearray().join((self._samples[voice], current_chunk))

    def _get_next_chunk(self, in_data, length, time_info, status):
        """Callback function to generate the next chunk to be played."""
        # this assumes 8-bit samples
        # if samples have run out, add silence
        samples = (
            _samp.ljust(length, b'\0') if len(_samp) < length else _samp[:length]
            for _samp in self._samples
        )
        # mix the samples
        mixed = bytearray(sum(_b) & 0xff for _b in zip(*samples))
        self._samples = [_samp[length:] for _samp in self._samples]
        return bytes(mixed), pyaudio.paContinue



"""
PC-BASIC - devicebase.py
Devices, Files and I/O operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import io
import os
import struct
import logging
from contextlib import contextmanager

from ...compat import iterchar
from ..base import error
from ..base.eascii import as_bytes as ea
from .. import values

def nullstream():
    return open(os.devnull, 'r+b')


# magic chars used by some devices to indicate file type
TYPE_TO_MAGIC = {b'B': b'\xFF', b'P': b'\xFE', b'M': b'\xFD'}
MAGIC_TO_TYPE = {b'\xFF': b'B', b'\xFE': b'P', b'\xFD': b'M'}



############################################################################
# Device classes
#
#  Some devices have a master file, where newly opened files inherit
#  width (and other?) settings from this file
#  For example, WIDTH "SCRN:", 40 works directly on the console,
#  whereas OPEN "SCRN:" FOR OUTPUT AS 1: WIDTH #1,23 works on the wrapper file
#  but does ot affect other files on SCRN: nor the console itself.
#  Likewise, WIDTH "LPT1:" works on LLIST etc and on lpt1 for the next time it's opened.


############################################################################

def parse_protocol_string(arg):
    """Retrieve protocol and options from argument."""
    if not arg:
        return None, u''
    argsplit = arg.split(u':', 1)
    if len(argsplit) == 1:
        addr, val = None, argsplit[0]
    else:
        addr, val = argsplit[0].upper(), u''.join(argsplit[1:])
    return addr, val


class NullDevice(object):
    """Null device (NUL) """

    def __init__(self):
        """Set up device."""

    def open(
            self, number, param, filetype, mode, access, lock,
            reclen, seg, offset, length, field
        ):
        """Open a file on the device."""
        return TextFileBase(nullstream(), filetype, mode)

    def close(self):
        """Close the device."""

    def available(self):
        """Device is available."""
        return True


class Device(object):
    """Device interface for master-file devices."""

    allowed_modes = b''

    def __init__(self):
        """Set up device."""
        self.device_file = None

    def open(
            self, number, param, filetype, mode, access, lock,
            reclen, seg, offset, length, field
        ):
        """Open a file on the device."""
        if not self.device_file:
            raise error.BASICError(error.DEVICE_UNAVAILABLE)
        if mode not in self.allowed_modes:
            raise error.BASICError(error.BAD_FILE_MODE)
        new_file = self.device_file.open_clone(filetype, mode, reclen)
        return new_file

    def close(self):
        """Close the device."""
        if self.device_file:
            self.device_file.close()

    def available(self):
        """Device is available."""
        return True


class SCRNDevice(Device):
    """Screen device (SCRN:) """

    allowed_modes = b'OR'

    def __init__(self, display, console):
        """Initialise screen device."""
        # open a master file on the screen
        Device.__init__(self)
        self.device_file = SCRNFile(display, console)

    def open(
            self, number, param, filetype, mode, access, lock,
            reclen, seg, offset, length, field
        ):
        """Open a file on the device."""
        new_file = Device.open(
            self, number, param, filetype, mode, access, lock,
            reclen, seg, offset, length, field
        )
        # SAVE "SCRN:" includes a magic byte
        new_file.write(TYPE_TO_MAGIC.get(filetype, b''))
        return new_file


class KYBDDevice(Device):
    """Keyboard device (KYBD:) """

    allowed_modes = b'IR'

    def __init__(self, keyboard, display):
        """Initialise keyboard device."""
        # open a master file on the keyboard
        Device.__init__(self)
        self.device_file = KYBDFile(keyboard, display)


#################################################################################
# file classes

# file interface:
#   __enter__(self)
#   __exit__(self, exc_type, exc_value, traceback)
#   close(self)
#   read(self, num=-1)
#   write(self, s)
#   filetype
#   mode


@contextmanager
def safe_io(err=error.DEVICE_IO_ERROR):
    """Catch and translate I/O errors."""
    try:
        yield
    except EnvironmentError as e:
        logging.warning('I/O error on stream access: %s', e)
        raise error.BASICError(err)


class RawFile(object):
    """File class for raw access to underlying stream."""

    def __init__(self, fhandle, filetype, mode):
        """Setup the basic properties of the file."""
        self._fhandle = fhandle
        self.filetype = filetype
        self.mode = mode.upper()

    def __enter__(self):
        """Context guard."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Context guard."""
        self.close()

    def close(self):
        """Close the file."""
        with safe_io():
            self._fhandle.close()

    def read(self, num=-1):
        """Read num chars. If num==-1, read all available."""
        with safe_io():
            return self._fhandle.read(num)

    def write(self, s):
        """Write string to file."""
        with safe_io():
            self._fhandle.write(s)

    def flush(self):
        """Stub for compatibility with file-like objects."""


#################################################################################
# Text file base

# text interface: file interface +
#   col
#   width
#   set_width(self, new_width=255)
#
#   read_line(self)
#   input_entry(self, typechar, allow_past_end)
#   write_line(self, s=b'')
#   eof(self)
#   lof(self)
#   loc(self)
#
#   internal use: read_one(), peek()
#   internal use: soft_sep

# TAB x09 is not whitespace for input#. NUL \x00 and LF \x0a are.
INPUT_WHITESPACE = b' \0\n'


class DeviceSettings(object):
    """Device-level width and column settings."""

    def __init__(self):
        """Setup the basic properties of the file."""
        self.width = 255
        self.col = 1

    def set_width(self, width):
        """Set file width."""
        self.width = width

    def close(self):
        """Close dummy device file."""


class TextFileBase(RawFile):
    """Base for text files on disk, KYBD file, field buffer."""

    # for INPUT# - numbers read from file can be separated by spaces too
    soft_sep = b' '

    def __init__(self, fhandle, filetype, mode):
        """Setup the basic properties of the file."""
        RawFile.__init__(self, fhandle, filetype, mode)
        # width=255 means line wrap
        self.width = 255
        self.col = 1
        # allow first char to be specified (e.g. already read)
        self._readahead = []
        self._current, self._previous = b'', b''

    # readable files

    def peek(self, num):
        """Return next num characters to be read; never returns more, fewer only at EOF."""
        to_read = num - len(self._readahead)
        if to_read > 0:
            with safe_io():
                self._readahead.extend(iterchar(self._fhandle.read(to_read)))
        return b''.join(self._readahead[:num])

    def read(self, num):
        """Read num characters."""
        output = self.peek(num)
        # check for \x1A - EOF char will actually stop further reading
        # (that's true in disk text files but not on COM devices)
        if b'\x1A' in output:
            output = output[:output.index(b'\x1A')]
        # drop read chars from buffer
        self._readahead = self._readahead[len(output):]
        if len(output) <= 1:
            self._previous = self._current
        else:
            self._previous = output[-2:]
        self._current = output[-1:]
        return output

    def read_one(self):
        """Read one character, converting device line ending to b'\r', EOF to b''."""
        # for use by input_entry() and read_line() only
        return self.read(1)

    def read_line(self):
        """\
            Read a single line until line break or 255 characters.
            Returns: (line, separator character)
            Separator character:
                b'\r' for device-standard line break (CR or CR LF)
                b'' at end of file
                None if line ended due to 255-char length limit
        """
        out = []
        while True:
            c = self.read_one()
            # don't check for CRLF on KYBD:, CAS:, etc.
            if not c or c == b'\r':
                break
            out.append(c)
            if len(out) == 255:
                c = b'\r' if self.peek(1) == b'\r' else None
                break
        return b''.join(out), c

    # writeable files

    def write(self, s, can_break=True):
        """Write the string s to the file, taking care of width settings."""
        assert isinstance(s, bytes)
        # only break lines at the start of a new string. width 255 means unlimited width
        s_width = 0
        newline = False
        # find width of first line in s
        for c in iterchar(s):
            if c in (b'\r', b'\n'):
                newline = True
                break
            if c >= b' ':
                # nonprinting characters including tabs are not counted for WIDTH
                s_width += 1
        if (
                can_break and self.width != 255 and self.col != 1 and
                self.col-1 + s_width > self.width and not newline
            ):
            self.write_line()
            self.col = 1
        for c in iterchar(s):
            # don't replace CR or LF with CRLF when writing to files
            if c == b'\r':
                self._fhandle.write(c)
                self.col = 1
            else:
                self._fhandle.write(c)
                # nonprinting characters including tabs are not counted for WIDTH
                if c >= b' ':
                    self.col += 1
                    # col-1 is a byte that wraps
                    if self.col == 257:
                        self.col = 1

    def write_line(self, s=b''):
        """Write string and follow with device-standard line break."""
        self.write(s + b'\r')

    # available for read & write (but not always useful)

    def set_width(self, new_width=255):
        """Set file width."""
        self.width = new_width

    def eof(self):
        """Check for end of file EOF."""
        # for EOF(i)
        if self.mode in (b'A', b'O'):
            return False
        return self.peek(1) in (b'', b'\x1a')


class InputMixin(object):
    """Support for INPUT#."""

    def _skip_whitespace(self, whitespace):
        """Skip spaces and line feeds and NUL; return last whitespace char """
        c = b''
        while True:
            next_char = self.peek(1)
            if not next_char or next_char not in whitespace:
                break
            # drop whitespace char
            c = self.read_one()
            # LF causes following CR to be dropped
            if c == b'\n' and self.peek(1) == b'\r':
                # LFCR: drop the CR, report as LF
                # on disk devices, this means LFCRLF is reported as LF
                self.read_one()
        return c

    def input_entry(self, typechar, allow_past_end, suppress_unquoted_linefeed=True):
        """Read a number or string entry for INPUT """
        word, blanks = b'', b''
        # fix readahead buffer (self.next_char)
        last = self._skip_whitespace(INPUT_WHITESPACE)
        # read first non-whitespace char
        c = self.read_one()
        # LF escapes quotes
        # may be true if last == '', hence "in ('\n', '\0')" not "in '\n0'"
        quoted = (c == b'"' and typechar == values.STR and last not in (b'\n', b'\0'))
        if quoted:
            c = self.read_one()
        # LF escapes end of file, return empty string
        if not c and not allow_past_end and last not in (b'\n', b'\0'):
            raise error.BASICError(error.INPUT_PAST_END)
        # we read the ending char before breaking the loop
        # this may raise FIELD OVERFLOW
        while c and not (
                (typechar != values.STR and c in self.soft_sep) or
                (c in b',\r' and not quoted)
            ):
            if c == b'"' and quoted:
                # whitespace after quote will be skipped below
                break
            elif suppress_unquoted_linefeed and (c == b'\n' and not quoted):
                # LF, LFCR are dropped entirely
                c = self.read_one()
                if c == b'\r':
                    c = self.read_one()
                continue
            elif c == b'\0':
                # NUL is dropped even within quotes
                pass
            elif c in INPUT_WHITESPACE and not quoted:
                # ignore whitespace in numbers, except soft separators
                # include internal whitespace in strings
                if typechar == values.STR:
                    blanks += c
            else:
                word += blanks + c
                blanks = b''
            if len(word) + len(blanks) >= 255:
                break
            if not quoted:
                c = self.read_one()
            else:
                # no CRLF replacement inside quotes.
                c = self.read(1)
        # if separator was a whitespace char or closing quote
        # skip trailing whitespace before any comma or hard separator
        if c and c in INPUT_WHITESPACE or (quoted and c == b'"'):
            self._skip_whitespace(b' ')
            if self.peek(1) in b',\r':
                c = self.read_one()
        # file position is at one past the separator char
        return word, c


#################################################################################
# Console INPUT


class InputTextFile(TextFileBase, InputMixin):
    """Handle INPUT from console."""

    # spaces do not separate numbers on console INPUT
    soft_sep = b''

    def __init__(self, line):
        """Initialise InputStream."""
        TextFileBase.__init__(self, io.BytesIO(line), b'D', b'I')


#################################################################################
# Console files


class RealTimeInputMixin(object):
    """Support for INPUT# on non-seekable KYBD and COM files."""

    def input_entry(self, typechar, allow_past_end):
        """Read a number or string entry from KYBD: or COMn: for INPUT#."""
        word, blanks = b'', b''
        c = self.read_one()
        # LF escapes quotes
        quoted = (c == b'"' and typechar == values.STR)
        if quoted:
            c = self.read_one()
        # LF escapes end of file, return empty string
        if not c and not allow_past_end:
            raise error.BASICError(error.INPUT_PAST_END)
        # on reading from a KYBD: file, control char replacement takes place
        # which means we need to use read_one() not read()
        parsing_trail = False
        while c and not (c in b',\r' and not quoted):
            if c == b'"' and quoted:
                parsing_trail = True
            elif c == b'\n' and not quoted:
                # LF, LFCR are dropped entirely
                c = self.read_one()
                if c == b'\r':
                    c = self.read_one()
                continue
            elif c == b'\0':
                # NUL is dropped even within quotes
                pass
            elif c in INPUT_WHITESPACE and not quoted:
                # ignore whitespace in numbers, except soft separators
                # include internal whitespace in strings
                if typechar == values.STR:
                    blanks += c
            else:
                word += blanks + c
                blanks = b''
            if len(word) + len(blanks) >= 255:
                break
            # there should be KYBD: control char replacement here even if quoted
            save_prev = self._previous
            c = self.read_one()
            if parsing_trail:
                if c not in INPUT_WHITESPACE:
                    # un-read the character if it's not a separator
                    if c not in (b',', b'\r'):
                        self._readahead.insert(0, c)
                        self._current, self._previous = self._previous, save_prev
                    break
            parsing_trail = parsing_trail or (typechar != values.STR and c == b' ')
        # file position is at one past the separator char
        return word, c


###############################################################################


# replace some eascii codes with control characters
KYBD_REPLACE = {
    ea.HOME: b'\xFF\x0B', ea.UP: b'\xFF\x1E', ea.PAGEUP: b'\xFE',
    ea.LEFT: b'\xFF\x1D', ea.RIGHT: b'\xFF\x1C', ea.END: b'\xFF\x0E',
    ea.DOWN: b'\xFF\x1F', ea.PAGEDOWN: b'\xFE',
    ea.DELETE: b'\xFF\x7F', ea.INSERT: b'\xFF\x12',
    ea.F1: b'', ea.F2: b'', ea.F3: b'', ea.F4: b'', ea.F5: b'',
    ea.F6: b'', ea.F7: b'', ea.F8: b'', ea.F9: b'', ea.F10: b'',
}

class KYBDFile(TextFileBase, RealTimeInputMixin):
    """KYBD device: keyboard."""

    col = 0

    def __init__(self, keyboard, display):
        """Initialise keyboard file."""
        TextFileBase.__init__(self, nullstream(), filetype=b'D', mode=b'I')
        # buffer for the separator character that broke the last INPUT# field
        # to be attached to the next
        self._keyboard = keyboard
        # screen needed for width settings on KYBD: master file
        self._display = display
        # on master-file devices, this is the master file.
        self._is_master = True

    def open_clone(self, filetype, mode, reclen=128):
        """Clone device file."""
        inst = KYBDFile(self._keyboard, self._display)
        inst.mode = mode
        inst.reclen = reclen
        inst.filetype = filetype
        inst._is_master = False
        return inst

    def peek(self, num):
        """Return only readahead buffer, no blocking peek."""
        return b''.join(self._readahead[:num])

    def read(self, num):
        """Read a number of characters (INPUT$)."""
        # take at most num chars out of readahead buffer (holds just one on KYBD but anyway)
        chars, self._readahead = b''.join(self._readahead[:num]), self._readahead[num:]
        # fill up the rest with actual keyboard reads
        while len(chars) < num:
            chars += b''.join(
                # note that INPUT$ on KYBD files replaces some eascii with NUL
                b'\0' if c in KYBD_REPLACE else c if len(c) == 1 else b''
                for c in self._keyboard.read_bytes_kybd_file(num-len(chars))
            )
        return chars

    def read_one(self):
        """Read a character with line ending replacement (INPUT and LINE INPUT)."""
        # take char out of readahead buffer, if present; blocking keyboard read otherwise
        if self._readahead:
            chars, self._readahead = b''.join(self._readahead[:1]), self._readahead[1:]
            return chars
        else:
            # note that we need string length, not list length
            # as read_bytes_kybd_file can return multi-byte eascii codes
            # blocking read
            return b''.join(
                # INPUT and LINE INPUT on KYBD files replace some eascii with control sequences
                KYBD_REPLACE.get(c, c)
                for c in self._keyboard.read_bytes_kybd_file(1)
            )

    # read_line: inherited from TextFileBase, this calls peek()

    def lof(self):
        """LOF for KYBD: is 1."""
        return 1

    def loc(self):
        """LOC for KYBD: is 0."""
        return 0

    def eof(self):
        """KYBD only EOF if ^Z is read."""
        if self.mode in (b'A', b'O'):
            return False
        # blocking peek
        return (self._keyboard.peek_byte_kybd_file() == b'\x1A')

    def set_width(self, new_width=255):
        """Setting width on KYBD device (not files) changes screen width."""
        if self._is_master:
            self._display.set_width(new_width)


###############################################################################

class SCRNFile(RawFile):
    """SCRN: file, allows writing to the screen as a text file."""

    def __init__(self, display, console):
        """Initialise screen file."""
        RawFile.__init__(self, nullstream(), filetype=b'D', mode=b'O')
        # need display object as WIDTH can change graphics mode
        self._display = display
        # screen member is public, needed by print_
        self.console = console
        self._width = self.console.width
        self._col = self.console.current_col
        # on master-file devices, this is the master file.
        self._is_master = True

    def open_clone(self, filetype, mode, reclen=128):
        """Clone screen file."""
        inst = SCRNFile(self._display, self.console)
        inst.mode = mode
        inst.reclen = reclen
        inst.filetype = filetype
        inst._is_master = False
        return inst

    def write(self, s, can_break=True):
        """Write string s to SCRN: """
        if not s:
            return
        # writes to SCRN files should *not* be echoed
        do_echo = self._is_master
        self._col = self.console.current_col
        # take column 80+overflow into account
        if self.console.overflow:
            self._col += 1
        # only break lines at the start of a new string. width 255 means unlimited width
        s_width = 0
        newline = False
        # find width of first line in s
        for c in iterchar(s):
            if c in (b'\r', b'\n'):
                newline = True
                break
            if c == b'\b':
                # for lpt1 and files, nonprinting chars are not counted in LPOS;
                # but chr$(8) will take a byte out of the buffer
                s_width -= 1
            elif c >= b' ':
                # nonprinting characters including tabs are not counted for WIDTH
                s_width += 1
        if can_break and (self.width != 255 and self.console.current_row != self.console.height
                and self.col != 1 and self.col-1 + s_width > self.width and not newline):
            self.console.write_line(do_echo=do_echo)
            self._col = 1
        cwidth = self.console.width
        output = []
        for c in iterchar(s):
            if self.width <= cwidth and self.col > self.width:
                self.console.write_line(b''.join(output), do_echo=do_echo)
                output = []
                self._col = 1
            if self.col <= cwidth or self.width <= cwidth:
                output.append(c)
            if c in (b'\n', b'\r'):
                self.console.write(b''.join(output), do_echo=do_echo)
                output = []
                self._col = 1
            else:
                self._col += 1
        self.console.write(b''.join(output), do_echo=do_echo)

    def write_line(self, inp=b''):
        """Write a string to the screen and follow by CR."""
        self.write(inp)
        self.console.write_line(do_echo=self._is_master)

    @property
    def col(self):
        """Return current (virtual) column position."""
        if self._is_master:
            return self.console.current_col
        else:
            return self._col

    @property
    def width(self):
        """Return (virtual) screen width."""
        if self._is_master:
            return self._display.mode.width
        else:
            return self._width

    def set_width(self, new_width=255):
        """Set (virtual) screen width."""
        if self._is_master:
            self._display.set_width(new_width)
        else:
            self._width = new_width

    def lof(self):
        """LOF: bad file mode."""
        raise error.BASICError(error.BAD_FILE_MODE)

    def loc(self):
        """LOC: bad file mode."""
        raise error.BASICError(error.BAD_FILE_MODE)

    def eof(self):
        """EOF: bad file mode."""
        raise error.BASICError(error.BAD_FILE_MODE)
